# -*- coding: utf-8 -*-

"""
Figure Canvas
=============

"""


# %% IMPORTS
# Built-in imports

# Package imports
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg


# All declaration
__all__ = ['FigureCanvas']


# %% CLASS DEFINITIONS
# Custom FigureCanvas class
class FigureCanvas(FigureCanvasQTAgg):
    pass



# -*- coding utf-8 -*-

"""
Data Table Widgets
==================

"""


# %% IMPORTS
# Import base modules
from . import data_table, headers, model, selection_model, view
from .data_table import *
from .headers import *
from .model import *
from .selection_model import *
from .view import *

# All declaration
__all__ = ['data_table', 'headers', 'model', 'selection_model', 'view']
__all__.extend(data_table.__all__)
__all__.extend(headers.__all__)
__all__.extend(model.__all__)
__all__.extend(selection_model.__all__)
__all__.extend(view.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Figure Manager
==============

"""


# %% IMPORTS
# Built-in imports

# Package imports
from matplotlib.backend_bases import FigureManagerBase


# All declaration
__all__ = ['FigureManager']


# %% CLASS DEFINITIONS
# Custom FigureManager class
class FigureManager(FigureManagerBase):
    pass



# -*- coding: utf-8 -*-

"""
Formatters Core
===============
Collects all the registered formatters for data tables into a single dict.

"""


# %% IMPORTS
# Built-in imports
from importlib import import_module
import os
from os import path

# Package imports
from sortedcontainers import SortedDict as sdict

# GuiPy imports
from guipy.config import register_file_format

# All declaration
__all__ = ['FORMATTERS', 'register_formatter']


# %% GLOBALS
# Define list of potential file formats
FORMATS_LIST = [
    "Windows Bitmap (*.bmp)",
    "Comma-Separated Values (*.csv)",
    "Encapsulated PostScript (*.eps)",
    "Flexible Image Transport System (*.fits)",
    "GuiPy Environment File (*.gpy)",
    "Hierarchical Data Format (*.hdf5 *.hdf4 *.hdf *.h5 *.h4 *.he5 *.he2)",
    "Joint Photographic Experts Group (*.jpg *.jpeg)",
    "NumPy Binary File (*.npy)",
    "NumPy Binary Archive (*.npz)",
    "Portable Document Format (*.pdf)",
    "PGF code for LaTeX (*.pgf)",
    "Portable Network Graphics (*.png)",
    "Postscript (*.ps)",
    "Python Script (*.py)",
    "Raw RGBA Bitmap (*.raw *.rgba)",
    "Scalable Vector Graphics (*.svg *.svgz)",
    "Portable Pixmap (*.ppm)",
    "Text Document (*.txt)",
    "X11 Bitmap (*.xbm)",
    "Excel File Format (*.xlsx *.xls)",
    "X11 Pixmap (*.xpm)"]

# Define dict of data table formatters
FORMATTERS = sdict()


# %% FUNCTION DEFINITIONS
# This function registers a data table formatter
def register_formatter(formatter_class):
    """
    Registers a provided data table formatter `formatter_class` for use in
    *GuiPy*.

    All data table formatters must be registered with this function in order to
    be used.

    Parameters
    ----------
    formatter_class : \
        :class:`~guipy.plugins.data_table.formatters.BaseFormatter` subclass
        The formatter class to use for formatting a data table.

    """

    # Initialize provided Formatter class
    formatter = formatter_class()

    # Register the file format that this formatter uses
    register_file_format(formatter.type, formatter.exts)

    # Register the formatter
    for ext in formatter.exts:
        FORMATTERS[ext] = formatter


# This function imports all pre-defined formatters and registers them
def _import_formatters():
    """
    Imports and registers all pre-defined data table formatters for use in
    *GuiPy*.

    """

    # Obtain the path to this directory
    dirpath = path.dirname(__file__)

    # Obtain a list of all files in this directory
    filenames = next(os.walk(dirpath))[2]

    # Remove __init__.py, base.py and core.py
    filenames.remove('__init__.py')
    filenames.remove('base.py')
    filenames.remove('core.py')

    # Loop over all modules and import their Formatter class
    for filename in filenames:
        # Obtain full module name
        modname = "%s.%s" % (__package__, filename[:-3])

        # Import this module
        mod = import_module(modname)

        # Register everything in __all__ as a formatter
        for prop in mod.__all__:
            formatter = getattr(mod, prop)
            register_formatter(formatter)


# Import all defined formatters
_import_formatters()



# -*- coding utf-8 -*-

"""
Figure Widgets
==============

"""


# %% IMPORTS
# Import base modules
from . import canvas, figure, manager, options, plot_entry, toolbar
from .canvas import *
from .figure import *
from .manager import *
from .options import *
from .plot_entry import *
from .toolbar import *

# Import subpackages
from . import types

# All declaration
__all__ = ['canvas', 'figure', 'manager', 'options', 'plot_entry',
           'toolbar', 'types']
__all__.extend(canvas.__all__)
__all__.extend(figure.__all__)
__all__.extend(manager.__all__)
__all__.extend(options.__all__)
__all__.extend(plot_entry.__all__)
__all__.extend(toolbar.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
GuiPy
=====
An easy plotting GUI for Python.

"""


# %% IMPORTS AND DECLARATIONS
# Import globals
from ._globals import *

# Import base modules and definitions
from .__version__ import __version__

# Import subpackages
from . import app, config, layouts, plugins, utils, widgets

# All declaration
__all__ = ['app', 'config', 'layouts', 'plugins', 'utils', 'widgets']

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Dock Widget
===========
Provides the base definition of the dock widgets used in *GuiPy*.

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import widgets as GW

# All declaration
__all__ = ['BaseDockWidget']


# %% CLASS DEFINITIONS
# Make base class for dock widgets
class BaseDockWidget(GW.QDockWidget):
    # Signals
    dockClosed = QC.Signal()

    # Override closeEvent to emit a signal whenever the dock is closed
    def closeEvent(self, *args, **kwargs):
        # Emit dockClosed signal
        self.dockClosed.emit()

        # Call super event
        super().closeEvent(*args, **kwargs)



# -*- coding: utf-8 -*-

"""
Base Formatters
===============

"""


# %% IMPORTS
# Built-in imports
import abc


# All declaration
__all__ = ['BaseFormatter']


# %% CLASS DEFINITIONS
# Define BaseFormatter abstract base class
class BaseFormatter(object, metaclass=abc.ABCMeta):
    """
    Provides an abstract base class definition that must be subclassed by all
    data table formatters.

    """

    # File type property (e.g., 'Portable Document Format')
    @property
    def type(self):
        # Return type if it is defined, or raise error if not
        if hasattr(self, 'TYPE'):
            return(self.TYPE)
        else:
            raise NotImplementedError("Class attribute 'TYPE' must be set by "
                                      "BaseFormatter subclass!")

    # File extensions property (e.g., ['.jpg', '.jpeg'])
    @property
    def exts(self):
        # Return exts if it is defined, or raise error if not
        if hasattr(self, 'EXTS'):
            return(self.EXTS)
        else:
            raise NotImplementedError("Class attribute 'EXTS' must be set by "
                                      "BaseFormatter subclass!")

    # Define exporter abstract method
    @abc.abstractmethod
    def exporter(self, data_table, filepath):
        """
        Exports the provided `data_table` to a %(ext)s-file.

        Parameters
        ----------
        data_table : :obj:`~guipy.plugins.data_table.widgets.DataTableWidget`\
            object
            The data table that must be exported.
        filepath : str
            The path to the file to be created.

        """

        # Raise NotImplementedError if only super() was called
        raise NotImplementedError("This method must be overridden in the "
                                  "BaseFormatter subclass!")

    # Define importer abstract method
    @abc.abstractmethod
    def importer(self, filepath, parent=None):
        """
        Imports a %(ext)s-file with the provided `filepath` as a
        :obj:`~pandas.DataFrame` object.

        Parameters
        ----------
        filepath : str
            The path to the %(ext)s-file.

        Optional
        --------
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent that will be maintaining the data.
            If *None*, no parent will be used.

        Returns
        -------
        data_table : :obj:`~pandas.DataFrame` object
            The data frame that contains all the read-in data.

        """

        # Raise NotImplementedError if only super() was called
        raise NotImplementedError("This method must be overridden in the "
                                  "BaseFormatter subclass!")



# -*- coding: utf-8 -*-

"""
LineEdits
=========

"""


# %% IMPORTS
# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW

# All declaration
__all__ = ['DualLineEdit', 'FloatLineEdit', 'IntLineEdit']


# %% CLASS DEFINITIONS
# Make class with two line-edits
class DualLineEdit(GW.DualBaseBox):
    """
    Defines the :class:`~DualLineEdit` class.

    """

    # Signals
    modified = QC.Signal([], [int, int], [int, float], [int, str],
                         [float, int], [float, float], [float, str],
                         [str, int], [str, float], [str, str])

    # Initialize the DualLineEdit class
    def __init__(self, types=(str, str), sep=None, parent=None):
        """
        Initialize an instance of the :class:`~DualLineEdit` class.

        Optional
        --------
        types : tuple of types ({int; float; str}). Default: (str, str)
            A tuple containing the type of each line-edit.
        sep : str or None. Default: None
            The string that must be used as a separator between the two
            line-edits. If *None*, no separator is used.
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this dual line-edit or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the dual line-edit
        self.init(types, sep)

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified.__getitem__(self.types))

    # This function creates the dual line-edit
    def init(self, types, sep):
        """
        Sets up the dual line-edit after it has been initialized.

        """

        # Make dict with different line-edits
        box_types = {
            float: FloatLineEdit,
            int: IntLineEdit,
            str: GW.QLineEdit}

        # Save provided types
        self.types = types

        # Create the box_layout
        box_layout = GL.QHBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create two line-edits with the provided types
        # LEFT
        left_box = box_types[types[0]]()
        box_layout.addWidget(left_box)
        self.left_box = left_box

        # RIGHT
        right_box = box_types[types[1]]()
        box_layout.addWidget(right_box)
        self.right_box = right_box

        # If sep is not None, create label and add it
        if sep is not None:
            sep_label = GW.QLabel(sep)
            sep_label.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
            box_layout.insertWidget(1, sep_label)

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        # Emit modified signal with proper types
        self.modified[self.types[0], self.types[1]].emit(
            *DualLineEdit.get_box_value(self))


# Make class for setting a number in a line-edit
class IntLineEdit(GW.QLineEdit):
    """
    Defines the :class:`~IntLineEdit` class.

    This class is used for creating a lineedit object that solely accepts
    integers.

    """

    # Signals
    modified = QC.Signal([float], [int])

    # Initialize the IntLineEdit class
    def __init__(self, parent=None):
        """
        Initialize an instance of the :class:`~IntLineEdit` class.

        Optional
        --------
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this number line-edit box or *None*
            for no parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the number line-edit box
        self.init()

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified.__getitem__(self.numtype))

    # This property returns the number type of this box
    @property
    def numtype(self):
        return(int)

    # This property returns the number getter of this box
    @property
    def num_getter(self):
        return(self.locale().toInt)

    # This property returns the proper validator to use
    def get_validator(self):
        return(QG.QIntValidator)

    # This function creates the number line-edit box
    def init(self):
        """
        Sets up the number line-edit box after it has been initialized.

        """

        # Obtain the proper validator
        validator = self.get_validator()(self)

        # Set the validator
        self.setValidator(validator)

        # Set initial value
        self.value = 0

    # Override focusInEvent to format text when it is triggered
    def focusInEvent(self, event):
        # Obtain a normal string version of the current number
        num = str(self.value)
        num = num.replace('.', self.locale().decimalPoint())

        # Set this as the current text
        self.setText(num)

        # Call and return super method
        return(super().focusInEvent(event))

    # Override focusOutEvent to format text when it is triggered
    def focusOutEvent(self, event):
        # Set current number in its formatted version
        self.set_box_value(self.num_getter(self.text())[0])

        # Call and return super method
        return(super().focusOutEvent(event))

    # This function calls the validator's setRange
    def setRange(self, bottom, top):
        self.validator().setRange(bottom, top)

    # This function calls the validator's setBottom
    def setBottom(self, bottom):
        self.validator().setBottom(bottom)

    # This function calls the validator's setTop
    def setTop(self, top):
        self.validator().setTop(top)

    # This function retrieves a value of this special box
    def get_box_value(self, *value_sig):
        """
        Returns the current number value of this line-edit box.

        Returns
        -------
        value : int, float
            The value contained in this line-edit bbox.

        """

        return(self.value)

    # This function sets the value of this special box
    def set_box_value(self, value, *value_sig):
        """
        Sets the current number value of this line-edit box to `value`.

        Parameters
        ----------
        value : int, float
            A value that must be set for this line-edit box.

        """

        # Save the current value
        cur_value = self.value

        # Save value
        self.value = value
        self.setText(self.locale().toString(value))

        # Emit modified signal if value was changed
        if(cur_value != self.value):
            self.default_modified_signal.emit(value)


# Make class for setting a number in a line-edit
class FloatLineEdit(IntLineEdit):
    """
    Defines the :class:`~FloatLineEdit` class.

    This class is used for creating a lineedit object that solely accepts
    floats.

    """

    # This property returns the number type of this box
    @property
    def numtype(self):
        return(float)

    # This property returns the number getter of this box
    @property
    def num_getter(self):
        return(self.locale().toDouble)

    # This property returns the proper validator to use
    def get_validator(self):
        return(QG.QDoubleValidator)



# -*- coding: utf-8 -*-

"""
Base Plot Types
===============

"""


# %% IMPORTS
# Package imports
from qtpy import QtCore as QC

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.plugins.figure.widgets.types.props import PLOT_PROPS

# All declaration
__all__ = ['BasePlotType']


# %% CLASS DEFINITIONS
# Define BasePlotType base class
class BasePlotType(GW.QWidget):
    """
    Provides a base class definition that must be subclassed by all figure plot
    types.

    Every plot type has a `name` and `prefix` associated with it; can be used
    for a specific `axis_type`; and a list of all `prop_names`.

    The `name` is the name/string used internally to identify this property.

    The `prefix` is the string used as a prefix for texts/names that relate to
    this plot type.

    The `display_name` is the text used in the types combobox.

    The `prop_names` are the names of the plot properties that this plot type
    uses. The 'Data' property is used by default and does not need to be
    provided.

    """

    # Define class attributes
    NAME = ""
    PREFIX = ""
    AXIS_TYPE = ""
    PROP_NAMES = []

    # Initialize this plot type
    def __init__(self, toolbar, parent=None, *args, **kwargs):
        # Save provided FigureToolbar object
        self.toolbar = toolbar
        self.options = toolbar.options_dialog
        self.data_table_plugin = toolbar.data_table_plugin
        self.figure = toolbar.canvas.figure
        self.axis = self.figure.gca()

        # Save the option getter
        self.get_option = self.options.get_option

        # Call super constructor
        super().__init__(parent)

        # Set up the plot entry box
        self.init(*args, **kwargs)

    # This function sets up the plot type
    def init(self):
        # Create layout for this line plot
        self.create_type_layout()

        # Save that currently no line exists
        self.plot = None

        # Attempt to update the plot to check if that does not raise errors
        self.update_plot()

    # This function creates the type layout
    def create_type_layout(self):
        # Create empty list of properties
        self.props = []

        # Create layout for this plot type
        layout = GL.QFormLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Connect signals
        self.options.refreshing_plots.connect(self.update_plot)

        # Loop over all required plot props
        for prop_name in self.PROP_NAMES:
            # Obtain the PlotProp class associated with this property
            plot_prop_class = PLOT_PROPS[prop_name]

            # Create a dictionary with all requirements of this property
            prop_kwargs = {req: getattr(self, req)
                           for req in plot_prop_class.REQUIREMENTS}

            # Initialize the property and add to layout
            # TODO: Create a collapsable QGroupBox widget
            prop_layout = plot_prop_class(**prop_kwargs)
            prop_group = GW.QGroupBox(prop_layout.DISPLAY_NAME)
            prop_group.setLayout(prop_layout)
            layout.addRow(prop_group)
            self.props.append(prop_layout)

            # Register all widgets in this property as instance attributes
            for widget_name, widget in prop_layout.widgets.items():
                setattr(self, widget_name, widget)

    # Define update_plot method
    @QC.Slot()
    def update_plot(self):
        """
        Draws and updates the current plot. This function must implement a
        check for whether updating is necessary/required.

        """

        raise NotImplementedError(self.__class__)

    # Define remove_plot method
    @QC.Slot()
    def remove_plot(self):
        """
        Removes the current plot.

        """

        raise NotImplementedError(self.__class__)

    # Override closeEvent to remove all plots and props
    def closeEvent(self, *args, **kwargs):
        # Remove the plots from the figure if they exist
        if self.plot is not None:
            self.remove_plot()

        # Close all props
        for prop in self.props:
            prop.close()

        # Call super event
        super().closeEvent(*args, **kwargs)



# -*- coding: utf-8 -*-

"""
Main Window
===========
Provides the definition of the main window of the *GuiPy* application.

"""


# %% IMPORTS
# Built-in imports
import platform
from struct import calcsize
import sys
from textwrap import dedent

# Package imports
import qtpy
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import __version__, APP_NAME, STR_TYPES
from guipy import plugins as GP, widgets as GW
from guipy.config import CONFIG
from guipy.widgets import create_exception_handler

# All declaration
__all__ = ['MainWindow']


# %% CLASS DEFINITIONS
# Define class for main window
# TODO: Write a lock-file system for GuiPy
class MainWindow(GW.QMainWindow):
    """
    Defines the :class:`~MainWindow` class for *GuiPy*.

    This class provides the main window for the GUI and combines all other
    widgets; layouts; and elements together.

    """

    # Signals
    exception = QC.Signal()

    # Initialize MainWindow class
    def __init__(self, *args, **kwargs):
        """
        Initialize an instance of the :class:`~MainWindow` class.

        """

        # Call super constructor
        super().__init__(*args, **kwargs)

        # Set up the main window
        self.init()

    # This function sets up the main window
    def init(self):
        """
        Sets up the main window after it has been initialized.

        This function is mainly responsible for initializing all other widgets
        that are required to make the GUI work, and connecting them together.

        """

        # Initialize the config manager
        CONFIG._init(self)

        # Make sure that the viewer is deleted when window is closed
        self.setAttribute(QC.Qt.WA_DeleteOnClose)

        # Set window title
        self.setWindowTitle(APP_NAME)

        # Disable the default context menu (right-click menu)
        self.setContextMenuPolicy(QC.Qt.NoContextMenu)

        # Create statusbar
        self.create_statusbar()

        # Create menus
        self.create_menus()

        # Create toolbars
        self.create_toolbars()

        # Add all required plugins
        self.add_plugins()

        # Add all remaining core actions
        self.add_core_actions()

        # Maximize the window
        self.setWindowState(QC.Qt.WindowMaximized)

        # Set the exception handler to an internal message window
        sys.excepthook = create_exception_handler(self)

        # Write the current GuiPy configuration to file
        CONFIG.write_config()

    # Override closeEvent to automatically close all plugins
    def closeEvent(self, *args, **kwargs):
        # Close all plugins in plugins dict
        for plugin in self.plugins.values():
            plugin.close()

        # Call super event
        super().closeEvent(*args, **kwargs)

    # This function creates the statusbar in the viewer
    def create_statusbar(self):
        """
        Creates the bottom-level statusbar of the main window, primarily used
        for displaying extended descriptions of actions.

        """

        # Obtain statusbar
        self.statusbar = self.statusBar()

    # This function adds a given menu to the top-level menu
    def addMenu(self, menu, parent_menu):
        """
        Adds a provided `menu` to the `parent_menu` and registers it.

        Parameters
        ----------
        menu : :obj:`~guipy.widgets.GW.QMenu` object
            The menu object that must be added to the top-level menubar.
        parent_menu : :obj:`~guipy.widgets.GW.QMenu` or \
            :obj:`~PyQt5.QtWidgets.QMenuBar` object
            The menu(bar) object to which the provided `menu` must be added.

        """

        # Add menu to the provided parent_menu
        parent_menu.addMenu(menu)

        # If parent_menu is the menubar, register menu as top-level
        if parent_menu is self.menuBar():
            self.menus[menu.name] = menu

        # Else, add it to the parent menu with the given name
        else:
            self.menus["%s/%s" % (parent_menu.name, menu.name)] = menu

    # This function creates the top-level menus in the viewer
    def create_menus(self):
        """
        Creates the menus in the top-level menubar of the main window.

        Other widgets can modify these menus to add additional actions to it.

        """

        # TOP-LEVEL MENUS
        # Make empty dict of top-level menus and actions
        self.menus = {}
        actions = {None: []}

        # Create file menu
        actions[None].append(GW.QMenu('File', '&File'))
        actions['File'] = []

        # Create view menu
        actions[None].append(GW.QMenu('View', '&View'))
        actions['View'] = []

        # Create help menu
        actions[None].append(GW.QMenu('Help', '&Help'))
        actions['Help'] = []

        # SUBMENUS
        # Add a New menu to file menu
        actions['File'].append(GW.QMenu('New', '&New...'))

        # Add a separator to file menu
        actions['File'].append(None)

        # Add a docks menu to view menu
        actions['View'].append(GW.QMenu('Docks', '&Docks'))

        # Add a separator to view menu
        actions['View'].append(None)

        # Add a toolbars menu to view menu
        actions['View'].append(GW.QMenu('Toolbars', '&Toolbars'))

        # Add all menus and actions
        self.add_menu_actions(actions)

    # Override addToolBar to register an action for toggling the toolbar
    def addToolBar(self, toolbar):
        """
        Adds a provided `toolbar` to the main window and registers it.

        Parameters
        ----------
        toolbar : :obj:`~guipy.widgets.GW.QToolBar` object
            The toolbar object that must be added to the main window.

        """

        # Call super method
        super().addToolBar(toolbar)

        # Register toolbar
        self.toolbars[toolbar.name] = toolbar

        # Add action for toggling the toolbar
        self.add_menu_actions({'View/Toolbars': [toolbar.toggleViewAction()]})

    # This function creates the toolbars in the viewer
    def create_toolbars(self):
        """
        Creates the toolbars of the main window.

        Other widgets can modify these toolbars to add additional actions to
        it.

        """

        # Make empty dict of toolbars and actions
        self.toolbars = {}
        actions = {}

        # FILE
        # Create file toolbar
        self.addToolBar(GW.QToolBar('File', 'File toolbar'))
        actions['File'] = []

        # ACTIONS
        # Add the File/New menu to the File toolbar
        actions['File'].append(self.menus['File/New'])

        # Add all actions
        self.add_toolbar_actions(actions)

    # This function adds all core actions to the menubar
    def add_core_actions(self):
        """
        Adds all core *GuiPy* actions to the top-level menubar, which are not
        related to any plugin or widget.

        """

        # Create dict of actions
        actions = {
            'File': [],
            'View': [],
            'Help': []}

        # FILE MENU
        # Add quit action to file menu
        quit_act = GW.QAction(
            self, '&Quit',
            shortcut=QG.QKeySequence.Quit,
            statustip="Quit %s" % (APP_NAME),
            triggered=self.close,
            role=GW.QAction.QuitRole)
        actions['File'].append(quit_act)

        # HELP MENU
        # Add config action to help menu
        config_act = GW.QAction(
            self, CONFIG.config_dialog.NAME,
            shortcut=QG.QKeySequence.Preferences,
            statustip="Open %s configuration window" % (APP_NAME),
            triggered=CONFIG.config_dialog,
            role=GW.QAction.PreferencesRole)
        actions['Help'].append(config_act)
        actions['Help'].append(None)

        # Add about action to help menu
        about_act = GW.QAction(
            self, '&About...',
            statustip="About %s" % (APP_NAME),
            triggered=self.about,
            role=GW.QAction.AboutRole)
        actions['Help'].append(about_act)

        # Add aboutQt action to help menu
        aboutqt_act = GW.QAction(
            self, 'About &Qt...',
            statustip="About Qt framework",
            triggered=QW.QApplication.aboutQt,
            role=GW.QAction.AboutQtRole)
        actions['Help'].append(aboutqt_act)

        # Add all actions to the top-level menu
        self.add_menu_actions(actions)

    # This function adds all plugins to the main window
    def add_plugins(self):
        """
        Adds all plugins to the main window.

        """

        # Initialize empty dict with plugins
        self.plugins = {}

        # Initialize and add all the plugins
        self.add_plugin(GP.DataTable)
        self.add_plugin(GP.Figure)

    # This function adds a plugin to the main window
    def add_plugin(self, plugin_class, **kwargs):
        """
        Initializes a provided `plugin_class` using its requirements and the
        given `kwargs`, adds it as a plugin to this main window and returns it.
        If `plugin_class` is associated with a widget, it will be added as a
        dock widget to this main window as well.

        """

        # Obtain all the plugin requirements of the provided plugin_class
        req_plugins = {req: self.plugins.get(req)
                       for req in plugin_class.REQ_PLUGINS}

        # Initialize provided plugin_class
        plugin_obj = plugin_class(req_plugins=req_plugins, parent=self,
                                  **kwargs)

        # Add plugin_obj to dict of all current plugins
        self.plugins[plugin_obj.TITLE] = plugin_obj

        # Add all menu actions of this plugin_obj to the proper menus
        self.add_menu_actions(plugin_obj.MENU_ACTIONS)

        # Add all toolbars of this plugin_obj to the main window
        for toolbar in plugin_obj.TOOLBARS:
            self.addToolBar(toolbar)

        # Add all toolbar actions of this plugin_obj to the proper toolbars
        self.add_toolbar_actions(plugin_obj.TOOLBAR_ACTIONS)

        # Add all status widgets of this plugin_obj to the statusbar
        for widget in plugin_obj.STATUS_WIDGETS:
            self.statusbar.addPermanentWidget(widget)

        # Check if this plugin_obj is an instance of BasePluginWidget
        if isinstance(plugin_obj, GP.BasePluginWidget):
            # If so, add it as a dock widget as well
            self.add_dockwidget(plugin_obj)

        # Return the created plugin_obj
        return(plugin_obj)

    # This function adds a dock widget to the main window
    def add_dockwidget(self, plugin_obj):
        """
        Adds a provided `plugin_obj` as a dock widget to this main window.

        """

        # Create a dock widget object
        dock_widget = GW.BaseDockWidget(plugin_obj.TITLE, self)

        # Add provided plugin as a widget to it
        dock_widget.setWidget(plugin_obj)

        # Add dock_widget to the main window
        self.addDockWidget(plugin_obj.LOCATION, dock_widget)

        # Add action for toggling the dock widget
        self.add_menu_actions({'View/Docks': [dock_widget.toggleViewAction()]})

    # This function adds all actions defined in a dict to the proper menus
    def add_menu_actions(self, actions_dict):
        """
        Adds all menu actions defined in the provided `actions_dict` to the
        associated menus.

        Parameters
        ----------
        actions_dict : dict of lists
            Dict containing the actions that need to be added to what menu.

        Dict variables
        --------------
        keyword : str or None
            String specifying to which menu all associated actions should be
            added. If *None*, the actions are added to the top-level menubar
            instead.
        actions : list of {None; :obj:`~guipy.widgets.GW.QMenu`; str; \
                           :obj:`~PyQt5.QtWidgets.QAction`; \
                           :obj:`~PyQt5.QtCore.Slot`}
            A list containing all actions that must be added to the menu
            specified with `keyword`.
            If *None*, a separator is added.
            If :obj:`~guipy.widgets.GW.QMenu` object, the given menu is added
            and registered.
            If str, a section with the given name is added.
            If :obj:`~PyQt5.QtWidgets.QAction`, the given action is added.
            If :obj:`~PyQt5.QtCore.Slot`, the given slot is connected to the
            menu's action's :attr:`~PyQt5.QtWidgets.QAction.triggered` signal.

        """

        # Loop over all menus in actions_dict
        for menu_name, actions in actions_dict.items():
            # Obtain the corresponding menu
            if menu_name is None:
                menu = self.menuBar()
            else:
                menu = self.menus[menu_name]

            # Loop over all actions that must be added to this menu
            for action in actions:
                # If action is None, add a menu separator
                if action is None:
                    menu.addSeparator()
                # Else, if action is a menu, add a new menu
                elif isinstance(action, GW.QMenu):
                    self.addMenu(action, menu)
                # Else, if action is a string, add a new section
                elif isinstance(action, STR_TYPES):
                    menu.addSection(action)
                # Else, if action is an action, add a new action
                elif isinstance(action, QW.QAction):
                    menu.addAction(action)
                # Else, set provided function as the default action
                else:
                    menu.menuAction().triggered.connect(action)

    # This function adds all actions defined in a dict to the proper toolbars
    def add_toolbar_actions(self, actions_dict):
        """
        Adds all toolbar actions defined in the provided `actions_dict` to the
        associated toolbars.

        Parameters
        ----------
        actions_dict : dict of lists
            Dict containing the actions that need to be added to what toolbar.

        Dict variables
        --------------
        keyword : str
            String specifying to which toolbar all associated actions should be
            added.
        actions : list of {None; :obj:`~guipy.widgets.GW.QMenu`; \
                           :obj:`~PyQt5.QtWidgets.QWidget`; \
                           :obj:`~PyQt5.QtWidgets.QAction`}
            A list containing all actions that must be added to the toolbar
            specified with `keyword`.
            If *None*, a separator is added.
            If :obj:`~guipy.widgets.GW.QMenu` object, the action of the given
            menu is added.
            If :obj:`~PyQt5.QtWidgets.QWidget`, the given widget is added.
            If :obj:`~PyQt5.QtWidgets.QAction`, the given action is added.

        """

        # Loop over all toolbars in actions_dict
        for toolbar_name, actions in actions_dict.items():
            # Obtain the corresponding toolbar
            toolbar = self.toolbars[toolbar_name]

            # Loop over all actions that must be added to this toolbar
            for action in actions:
                # If action is None, add a toolbar separator
                if action is None:
                    toolbar.addSeparator()
                # Else, if action is a menu, add a new menu
                elif isinstance(action, GW.QMenu):
                    toolbar.addMenu(action)
                # Else, if action is a widget, add a new widget
                elif isinstance(action, QW.QWidget):
                    toolbar.addWidget(action)
                # Else, add the action to the toolbar
                else:
                    toolbar.addAction(action)

    # This function creates a message box with the 'about' information
    @QC.Slot()
    def about(self):
        """
        Displays a small section with information about *GuiPy*.

        """

        # Make shortcuts for certain links
        github_repo = "https://github.com/1313e/GuiPy"

        # Create the text for the 'about' dialog
        text = dedent(r"""
            <b><a href="{github}">{name}</a> v{version}</b><br>
            Copyright &copy; 2019-2020 Ellert van der Velden<br>
            Distributed under the
            <a href="{github}/raw/master/LICENSE">BSD-3 License</a>.<br>
            <br>
            {os_name} {os_release} | Python {python_version} {memsize}-bit |
            Qt {Qt_version} | PyQt5 {PyQt5_version}
            """.format(name=APP_NAME,
                       version=__version__,
                       github=github_repo,
                       os_name=platform.system(),
                       os_release=platform.release(),
                       python_version=platform.python_version(),
                       memsize=calcsize('P')*8,
                       Qt_version=qtpy.QT_VERSION,
                       PyQt5_version=qtpy.PYQT_VERSION))

        # Create the 'about' dialog
        GW.QMessageBox.about(self, "About %s" % (APP_NAME), text)



# -*- coding: utf-8 -*-

"""
Data Table Plugin
=================

"""


# %% IMPORTS
# Built-in imports
from os import path

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, plugins as GP, widgets as GW
from guipy.config import FILE_FILTERS
from guipy.plugins.data_table.formatters import FORMATTERS
from guipy.plugins.data_table.widgets import DataTableWidget
from guipy.widgets import set_box_value

# All declaration
__all__ = ['DataTable']


# %% CLASS DEFINITIONS
# Define class for the DataTable plugin
class DataTable(GP.BasePluginWidget):
    # Properties
    TITLE = "Data table"
    LOCATION = QC.Qt.LeftDockWidgetArea

    # Initialize DataTable plugin
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Set up the data table plugin
        self.init()

    # This function sets up the data table plugin
    def init(self):
        # Create a layout
        layout = GL.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Create a tab widget
        tab_widget = GW.QTabWidget(browse_tabs=True)
        tab_widget.setTabBar(GW.EditableTabBar())
        tab_widget.setMovable(True)
        tab_widget.setTabsClosable(True)

        # Connect tab widget signals
        tab_widget.tabCloseRequested.connect(self.close_tab)

        # Add tab widget to layout
        self.tab_widget = tab_widget
        layout.addWidget(self.tab_widget)

        # Add all actions to the proper menus and toolbars
        self.add_actions()

        # Add a tab to the plugin
        self.add_tab()

    # Override closeEvent to do automatic clean-up
    def closeEvent(self, *args, **kwargs):
        # Block all signals emitted by the tab widget while removing tabs
        self.tab_widget.blockSignals(True)

        # Close all tabs
        for index in reversed(range(self.tab_widget.count())):
            self.close_tab(index)

        # Call super event
        super().closeEvent(*args, **kwargs)

    # This function adds all associated actions to the menus and toolbars
    def add_actions(self):
        # Initialize empty action lists for this plugin
        self.MENU_ACTIONS = {
            **GP.BasePluginWidget.MENU_ACTIONS,
            'File': [],
            'File/New': []}
        self.TOOLBAR_ACTIONS = {
            **GP.BasePluginWidget.TOOLBAR_ACTIONS,
            'File': []}

        # Add new tab action to file/new menu
        new_tab_act = GW.QAction(
            self, 'Data &table',
            shortcut=QC.Qt.CTRL + QC.Qt.Key_T,
            tooltip="New data table",
            triggered=self.add_tab,
            role=GW.QAction.ApplicationSpecificRole)
        self.MENU_ACTIONS['File/New'].append(new_tab_act)
        self.MENU_ACTIONS['File/New'].append(self.add_tab)

        # Add open tabs action to file menu/toolbar
        open_tabs_act = GW.QAction(
            self, '&Open...',
            shortcut=QC.Qt.CTRL + QC.Qt.Key_O,
            tooltip="Open data table",
            triggered=self.open_tabs,
            role=GW.QAction.ApplicationSpecificRole)
        open_tabs_act.setEnabled(False)
        self.MENU_ACTIONS['File'].append(open_tabs_act)
        self.TOOLBAR_ACTIONS['File'].append(open_tabs_act)

        # Add import tabs action to file menu/toolbar
        import_tabs_act = GW.QAction(
            self, '&Import...',
            shortcut=QC.Qt.CTRL + QC.Qt.Key_I,
            tooltip="Import data tables",
            triggered=self.import_tabs,
            role=GW.QAction.ApplicationSpecificRole)
        self.MENU_ACTIONS['File'].append(import_tabs_act)
        self.TOOLBAR_ACTIONS['File'].append(import_tabs_act)

        # Add separator to file menu
        self.MENU_ACTIONS['File'].append(None)

        # Add save tab action to file menu/toolbar
        save_tab_act = GW.QAction(
            self, '&Save',
            shortcut=QC.Qt.CTRL + QC.Qt.Key_S,
            tooltip="Save current data table",
            triggered=self.save_tab,
            role=GW.QAction.ApplicationSpecificRole)
        save_tab_act.setEnabled(False)
        self.MENU_ACTIONS['File'].append(save_tab_act)
        self.TOOLBAR_ACTIONS['File'].append(save_tab_act)

        # Add save_as tab action to file menu
        save_as_tab_act = GW.QAction(
            self, 'Save &as...',
            shortcut=QC.Qt.CTRL + QC.Qt.SHIFT + QC.Qt.Key_S,
            tooltip="Save current data table as...",
            triggered=self.save_as_tab,
            role=GW.QAction.ApplicationSpecificRole)
        save_as_tab_act.setEnabled(False)
        self.MENU_ACTIONS['File'].append(save_as_tab_act)

        # Add save_all tab action to file menu/toolbar
        save_all_tabs_act = GW.QAction(
            self, 'Sav&e all',
            shortcut=QC.Qt.CTRL + QC.Qt.ALT + QC.Qt.Key_S,
            tooltip="Save all data tables",
            triggered=self.save_all_tabs,
            role=GW.QAction.ApplicationSpecificRole)
        save_all_tabs_act.setEnabled(False)
        self.MENU_ACTIONS['File'].append(save_all_tabs_act)
        self.TOOLBAR_ACTIONS['File'].append(save_all_tabs_act)

        # Add export tab action to file menu/toolbar
        export_tab_act = GW.QAction(
            self, '&Export...',
            shortcut=QC.Qt.CTRL + QC.Qt.Key_E,
            tooltip="Export current data table",
            triggered=self.export_tab,
            role=GW.QAction.ApplicationSpecificRole)
        self.MENU_ACTIONS['File'].append(export_tab_act)
        self.TOOLBAR_ACTIONS['File'].append(export_tab_act)

        # Add separator to file menu
        self.MENU_ACTIONS['File'].append(None)

    # This function adds a new data table widget
    @QC.Slot()
    def add_tab(self, name=None, import_func=None):
        # Create a new DataTableWidget
        data_table = DataTableWidget(self, import_func)

        # If name is None, set it to default
        if name is None:
            name = "table_%i" % (self.tab_widget.count())

        # Add data_table to the tab widget
        index = self.tab_widget.addTab(data_table, name)

        # Switch focus to the new tab
        set_box_value(self.tab_widget, index)

    # This function closes a data table widget
    # TODO: Warn the user about closing a tab if it has unsaved changes
    @QC.Slot(int)
    def close_tab(self, index):
        # Obtain the DataTableWidget object associated with this index
        data_table = self.dataTable(index)

        # Close this data_table
        data_table.close()

        # Remove this data_table from the tab widget
        self.tab_widget.removeTab(index)

    # This function opens a data table widget
    @QC.Slot()
    def open_tabs(self):
        pass

    # This function imports a data table widget
    @QC.Slot()
    def import_tabs(self):
        # Open the file opening system
        filepaths, _ = GW.getOpenFileNames(
            parent=self,
            caption="Import data tables",
            filters=FORMATTERS.keys())

        # Loop over filepaths and make a tab for every entry
        for filepath in filepaths:
            # Obtain the name and extension of this data table
            name, ext = path.splitext(path.basename(filepath))

            # Add a new tab
            self.add_tab(name, lambda x: FORMATTERS[ext].importer(filepath, x))

    # This function saves a data table widget
    @QC.Slot()
    def save_tab(self):
        pass

    # This function saves a data table widget with chosen name
    @QC.Slot()
    def save_as_tab(self):
        pass

    # This function saves all data table widgets
    @QC.Slot()
    def save_all_tabs(self):
        pass

    # This function exports a data table
    @QC.Slot()
    def export_tab(self):
        # Get name of this data table
        name = self.tabName()

        # Open the file saving system
        filepath, selected_filter = GW.getSaveFileName(
            parent=self,
            caption="Export data table %r to..." % (name),
            basedir=name+'.npz',
            filters=FORMATTERS.keys())

        # If filepath is not empty, export data table
        if filepath:
            # Obtain the ext of the filepath
            ext = path.splitext(filepath)[1]

            # If ext is empty, check what filter was used
            if not ext:
                # If "All (Supported) Files" was not used, get ext from filter
                if not selected_filter.startswith("All "):
                    ext = FILE_FILTERS[selected_filter]
                # Else, set it to '.npz'
                else:
                    ext = '.npz'

                # Add extension to filepath
                filepath += ext

            # Export data table
            FORMATTERS[ext].exporter(self.dataTable(), filepath)

    # This function returns the data table belonging to a specified int
    @QC.Slot(int)
    def dataTable(self, index=None):
        """
        Returns the :obj:`~DataTableWidget` object that belongs to the table
        with the provided tab `index`.

        Optional
        --------
        index : int or None. Default: None
            If int, the index of the tab whose table is requested.
            If *None*, the current table is requested.

        Returns
        -------
        data_table : :obj:`~DataTableWidget` object
            The data table that belongs to the tab specified by the
            provided `index`.

        """

        # If index is None, return current data table widget
        if index is None:
            index = self.tab_widget.currentIndex()

        # Return the data table widget with the provided index
        return(self.tab_widget.widget(index))

    # This function returns the name of the tab belonging to the specified int
    @QC.Slot(int)
    def tabName(self, index=None):
        """
        Returns the name of the tab with the provided tab `index`.

        Optional
        --------
        index : int or None. Default: None
            If int, the index of the tab whose name is requested.
            If *None*, the current tab name is requested.

        Returns
        -------
        name : str
            The name of the tab specified by the provided `index`.

        """

        # If index is None, return current tab name
        if index is None:
            index = self.tab_widget.currentIndex()

        # Return the name of the tab with the provided index
        return(self.tab_widget.tabText(index))



# -*- coding: utf-8 -*-

"""
Base Widgets
============
Provides a collection of custom :class:`~PyQt5.QtWidgets.QWidget` base classes
that allow for certain widgets to be standardized.

"""


# %% IMPORTS
# Built-in imports
from sys import platform

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports

# All declaration
__all__ = ['QAbstractButton', 'QAction', 'QCheckBox', 'QComboBox', 'QDialog',
           'QDockWidget', 'QDoubleSpinBox', 'QFileDialog', 'QFontComboBox',
           'QGroupBox', 'QHeaderView', 'QLabel', 'QLineEdit', 'QListView',
           'QListWidget', 'QMainWindow', 'QMenu', 'QMessageBox', 'QPushButton',
           'QRadioButton', 'QSpinBox', 'QSplitter', 'QScrollArea',
           'QStackedWidget', 'QTabBar', 'QTableView', 'QTabWidget',
           'QTextEdit', 'QToolBar', 'QToolButton', 'QToolTip', 'QWidget']


# %% BASE CLASS DEFINITION
# Make subclass of QWidget to provide certain functionality to all widgets
class QWidget(QW.QWidget):
    """
    Defines the :class:`~QWidget` class.

    This class provides default settings and extra options for the
    :class:`~PyQt5.QtWidgets.QWidget` class and is inherited by all other
    custom :class:`~PyQt5.QtWidgets.QWidget` classes.

    """

    # Initialize QWidget
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Retrieve certain methods from parent
        self.get_parent_methods()

    # This function retrieves a set of methods from the parent if possible
    def get_parent_methods(self):
        # Obtain parent widget
        parent = self.parentWidget()

        # If this widget has a parent, retrieve a few methods if possible
        if parent is not None:
            # Retrieve the 'get_option' method if it exists
            if(not hasattr(self, 'get_option') and
               hasattr(parent, 'get_option')):
                self.get_option = parent.get_option

    # Override setStatusTip to auto translate
    def setStatusTip(self, text):
        super().setStatusTip(text)

    # Override setToolTip to auto translate
    def setToolTip(self, text):
        super().setToolTip(text)

    # Override childEvent to add 'get_option' if it exists to all children
    def childEvent(self, event):
        """
        Special :meth:`~PyQt5.QtCore.QObject.childEvent` event that
        automatically calls the :meth:`~get_parent_methods` method on any
        widget that becomes a child of this widget.

        """

        # If this event involved a child being added, check child object
        if(event.type() == QC.QEvent.ChildAdded):
            # Obtain child object
            child = event.child()

            # If this child has the 'get_parent_methods' method, call it
            if hasattr(child, 'get_parent_methods'):
                child.get_parent_methods()

        # Call and return super method
        return(super().childEvent(event))

    # Override setLocale to also set it for all children
    def setLocale(self, locale):
        # Set locale for this object
        super().setLocale(locale)

        # Also set this locale for all children that are widgets
        for child in self.children():
            if isinstance(child, QWidget):
                child.setLocale(locale)


# %% CLASS DEFINITIONS
# Create custom QAbstractButton
class QAbstractButton(QW.QAbstractButton, QWidget):
    # Override constructor to set some default settings
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Use default settings
        self.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)


# Make subclass of QW.QAction that automatically sets details based on status
class QAction(QW.QAction):
    """
    Defines the :class:`~QAction` class.

    This class provides default settings and extra options for the
    :class:`~PyQt5.QtWidgets.QAction` class.

    """

    # Override constructor
    def __init__(self, parent, text, *, shortcut=None, tooltip=None,
                 statustip=None, icon=None, triggered=None, toggled=None,
                 role=None):
        """
        Initializes the :class:`~QAction` class.

        Parameters
        ----------
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None
            The parent widget for this dialog or *None* for no parent.
        text : str
            The label that this action must have.

        Optional
        --------
        shortcut : :obj:`~PyQt5.QtGui.QKeySequence` or None. Default: None
            The key sequence that must be set as the shortcut for this action.
            If *None*, no shortcut will be set.
        tooltip : str or None. Default: None
            The text that must be set as the tooltip for this action.
            If *None*, the tooltip is set to `text`.
            If `shortcut` is not *None*, the tooltip will also include the
            shortcut.
        statustip : str or None. Default: None
            The text that must be set as the statustip for this action.
            If *None*, the statustip is set to `tooltip`.
        icon : :obj:`~PyQt5.QtGui.QIcon` object or None. Default: None
            The icon that must be set as the icon for this action.
            If *None*, no icon will be set.
        triggered : function or None. Default: None
            The Qt slot function that must be called whenever this action is
            triggered.
            If *None*, no slot will be connected to this action's `triggered`
            signal.
        toggled : function or None. Default: None
            The Qt slot function that must be called whenever this action is
            toggled.
            If *None*, no slot will be connected to this action's `toggled`
            signal.
        role : :obj:`~PyQt5.QtWidgets.QAction.MenuRole` object or None. \
            Default: None
            The menu role that must be set as the role of this action.
            If *None*, it is set to :obj:`~PyQt5.QtWidgets.NoRole`.

        """

        # Call super constructor
        if icon is None:
            super().__init__(text, parent)
        else:
            super().__init__(icon, text, parent)

        # Set all the details
        self.setDetails(shortcut=shortcut,
                        tooltip=tooltip,
                        statustip=statustip)

        # Set the triggered signal
        if triggered is not None:
            self.triggered.connect(triggered)

        # Set the toggled signal
        if toggled is not None:
            self.toggled.connect(toggled)
            self.setCheckable(True)

        # Set the action menu role
        self.setMenuRole(self.NoRole if role is None else role)

    # Make new method that automatically sets Shortcut, ToolTip and StatusTip
    def setDetails(self, *, shortcut=None, tooltip=None, statustip=None):
        """
        Uses the provided `shortcut`; `tooltip`; and `statustip` to set the
        details of this action.

        Parameters
        ----------
        shortcut : :obj:`~PyQt5.QtGui.QKeySequence` or None. Default: None
            The key sequence that must be set as the shortcut for this action.
            If *None*, no shortcut will be set.
        tooltip : str or None. Default: None
            The text that must be set as the tooltip for this action.
            If *None*, the tooltip is set to `text`.
            If `shortcut` is not *None*, the tooltip will also include the
            shortcut.
        statustip : str or None. Default: None
            The text that must be set as the statustip for this action.
            If *None*, the statustip is set to `tooltip`.

        """

        # If shortcut is not None, set it
        if shortcut is not None:
            super().setShortcut(shortcut)
            shortcut = self.shortcut().toString()

        # If tooltip is None, its base is set to the action's name
        if tooltip is None:
            base_tooltip = self.text().replace('&', '')
            tooltip = base_tooltip
        # Else, provided tooltip is used as the base
        else:
            base_tooltip = tooltip

        # If shortcut is not None, add it to the tooltip
        if shortcut is not None:
            tooltip = "%s (%s)" % (base_tooltip, shortcut)

        # Set tooltip
        super().setToolTip(tooltip)

        # If statustip is None, it is set to base_tooltip
        if statustip is None:
            statustip = base_tooltip

        # Set statustip
        super().setStatusTip(statustip)

    # Override setShortcut to raise an error when used
    def setShortcut(self, *args, **kwargs):
        raise AttributeError("Using this method is not allowed! Use "
                             "'setDetails()' instead!")

    # Override setToolTip to raise an error when used
    def setToolTip(self, *args, **kwargs):
        raise AttributeError("Using this method is not allowed! Use "
                             "'setDetails()' instead!")

    # Override setStatusTip to raise an error when used
    def setStatusTip(self, *args, **kwargs):
        raise AttributeError("Using this method is not allowed! Use "
                             "'setDetails()' instead!")


# Create custom QCheckBox
class QCheckBox(QW.QCheckBox, QAbstractButton):
    pass


# Create custom combobox class with more signals
class QComboBox(QW.QComboBox, QWidget):
    """
    Defines the :class:`~QComboBox` class.

    This class provides default settings and extra options for the
    :class:`~PyQt5.QtWidgets.QComboBox` class.

    """

    # Signals
    popup_shown = QC.Signal([int], [str])
    popup_hidden = QC.Signal([int], [str])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setSizeAdjustPolicy(self.AdjustToContents)

    # Override the showPopup to emit a signal whenever it is triggered
    def showPopup(self, *args, **kwargs):
        self.popup_shown[int].emit(self.currentIndex())
        self.popup_shown[str].emit(self.currentText())
        return(super().showPopup(*args, **kwargs))

    # Override the hidePopup to emit a signal whenever it is triggered.
    def hidePopup(self, *args, **kwargs):
        self.popup_hidden[int].emit(self.currentIndex())
        self.popup_hidden[str].emit(self.currentText())
        return(super().hidePopup(*args, **kwargs))


# Create custom QFontComboBox class
class QFontComboBox(QW.QFontComboBox, QWidget):
    # Signals
    popup_shown = QC.Signal([int], [str])
    popup_hidden = QC.Signal([int], [str])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setSizeAdjustPolicy(self.AdjustToContents)

    # Override the showPopup to emit a signal whenever it is triggered
    def showPopup(self, *args, **kwargs):
        self.popup_shown[int].emit(self.currentIndex())
        self.popup_shown[str].emit(self.currentText())
        return(super().showPopup(*args, **kwargs))

    # Override the hidePopup to emit a signal whenever it is triggered.
    def hidePopup(self, *args, **kwargs):
        self.popup_hidden[int].emit(self.currentIndex())
        self.popup_hidden[str].emit(self.currentText())
        return(super().hidePopup(*args, **kwargs))


# Create custom QDialog
class QDialog(QW.QDialog, QWidget):
    pass


# Create custom QDockWidget
# TODO: Add a context menu button to all dockwidgets by default?
class QDockWidget(QW.QDockWidget, QWidget):
    pass


# Create custom QFileDialog class
class QFileDialog(QW.QFileDialog, QDialog):
    pass


# Create custom QGroupBox class
class QGroupBox(QW.QGroupBox, QWidget):
    pass


# Create custom QHeaderView class
class QHeaderView(QW.QHeaderView, QWidget):
    pass


# Create custom QAbstractSpinBox that automatically sets some properties
class QAbstractSpinBox(QW.QAbstractSpinBox, QWidget):
    """
    Defines the :class:`~QAbstractSpinBox` class.

    This class provides default settings and extra options for the
    :class:`~PyQt5.QtWidgets.QAbstractSpinBox` class.

    """

    # Override constructor to set some default settings
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Use default settings
        self.setStepType(self.AdaptiveDecimalStepType)
        self.setAccelerated(True)
        self.setGroupSeparatorShown(True)
        self.setStyleSheet(
            """
            QAbstractSpinBox {{
                margin: {0}px 0px {0}px 0px;
                max-height: 24px;}}
            """.format("-1" if platform.startswith('linux') else '0'))

    # Auto translate any special value text that is set
    def setSpecialValueText(self, text):
        super().setSpecialValueText(text)


# Create custom QDoubleSpinBox
class QDoubleSpinBox(QW.QDoubleSpinBox, QAbstractSpinBox):
    pass


# Create custom QSpinBox
class QSpinBox(QW.QSpinBox, QAbstractSpinBox):
    pass


# Create custom QLabel class with more signals
class QLabel(QW.QLabel, QWidget):
    """
    Defines the :class:`~QLabel` class.

    This class provides default settings and extra options for the
    :class:`~PyQt5.QtWidgets.QLabel` class.

    """

    # Signals
    contentsChanged = QC.Signal([str], [QG.QMovie], [QG.QPicture],
                                [QG.QPixmap])
    mousePressed = QC.Signal()

    def __init__(self, text=None, *args, **kwargs):
        # Call super constructor
        if text is None:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(text, *args, **kwargs)

        # Set some settings
        self.setWordWrap(True)
        self.setOpenExternalLinks(True)

    # Override the mousePressEvent to emit a signal whenever it is triggered
    def mousePressEvent(self, event):
        self.mousePressed.emit()
        event.accept()

    # Override setMovie to emit a signal whenever it is called
    def setMovie(self, movie):
        super().setMovie(movie)
        self.contentsChanged[QG.QMovie].emit(self.movie())

    # Override setNum to emit a signal whenever it is called
    def setNum(self, num):
        super().setNum(num)
        self.contentsChanged[str].emit(self.text())

    # Override setPicture to emit a signal whenever it is called
    def setPicture(self, picture):
        super().setPicture(picture)
        self.contentsChanged[QG.QPicture].emit(self.picture())

    # Override setPixmap to emit a signal whenever it is called
    def setPixmap(self, pixmap):
        super().setPixmap(pixmap)
        self.contentsChanged[QG.QPixmap].emit(self.pixmap())

    # Override setText to emit a signal whenever it is called
    def setText(self, text):
        super().setText(text)
        self.contentsChanged[str].emit(self.text())


# Create custom QLineEdit class
class QLineEdit(QW.QLineEdit, QWidget):
    pass


# Create custom QListView class
class QListView(QW.QListView, QWidget):
    pass


# Create custom QListWidget class
class QListWidget(QW.QListWidget, QWidget):
    pass


# Create custom QMainWindow class
class QMainWindow(QW.QMainWindow, QWidget):
    pass


# Create custom QMenu class
class QMenu(QW.QMenu, QWidget):
    """
    Defines the :class:`~QMenu` class.

    This class provides default settings and extra options for the
    :class:`~PyQt5.QtWidgets.QMenu` class.

    """

    def __init__(self, name, title=None, parent=None, *, tooltip=None,
                 statustip=None):
        # Save name
        self.name = name

        # If title is None, set it to name
        if title is None:
            title = name

        # Call super constructor
        super().__init__(title, parent)

        # Set all the details
        self.setDetails(tooltip=tooltip,
                        statustip=statustip)

    # Make new method that automatically sets ToolTip and StatusTip
    def setDetails(self, *, tooltip=None, statustip=None):
        """
        Uses the provided `tooltip` and `statustip` to set the details of this
        menu action.

        Parameters
        ----------
        tooltip : str or None. Default: None
            The text that must be set as the tooltip for this menu.
            If *None*, the tooltip is set to `title`.
        statustip : str or None. Default: None
            The text that must be set as the statustip for this menu.
            If *None*, the statustip is set to `tooltip`.

        """

        # Obtain the action that triggers this menu
        menu_action = self.menuAction()

        # If tooltip is None, it is set to the menu's name
        if tooltip is None:
            tooltip = self.title().replace('&', '')

        # Set tooltip
        menu_action.setToolTip(tooltip)

        # If statustip is None, it is set to tooltip
        if statustip is None:
            statustip = tooltip

        # Set statustip
        menu_action.setStatusTip(statustip)

    # Override setToolTip to raise an error when used
    def setToolTip(self, *args, **kwargs):
        raise AttributeError("Using this method is not allowed! Use "
                             "'setDetails()' instead!")

    # Override setStatusTip to raise an error when used
    def setStatusTip(self, *args, **kwargs):
        raise AttributeError("Using this method is not allowed! Use "
                             "'setDetails()' instead!")

    # Override addSection to automatically translate the given section name
    def addSection(self, text, icon=None):
        # Call super method
        if icon is None:
            return(super().addSection(text))
        else:
            return(super().addSection(icon, text))


# Create custom QMessageBox class
class QMessageBox(QW.QMessageBox, QDialog):
    pass


# Create custom QPushButton class
class QPushButton(QW.QPushButton, QAbstractButton):
    pass


# Create custom QRadioButton class
class QRadioButton(QW.QRadioButton, QAbstractButton):
    pass


# Create custom QScrollArea class
class QScrollArea(QW.QScrollArea, QWidget):
    pass


# Create custom QSplitter class
class QSplitter(QW.QSplitter, QWidget):
    pass


# Create custom QStackedWidget class
class QStackedWidget(QW.QStackedWidget, QWidget):
    pass


# Create custom QTabBar class
class QTabBar(QW.QTabBar, QWidget):
    # Signals
    tabTextChanged = QC.Signal(int, str)

    # Override constructor to set some default settings
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Use default settings
        self.setElideMode(QC.Qt.ElideNone)

    # Override setTabText to emit a signal whenever it is called
    def setTabText(self, index, text):
        # Emit signal
        self.tabTextChanged.emit(index, text)

        # Call super method
        return(super().setTabText(index, text))


# Create custom QTableView class
class QTableView(QW.QTableView, QWidget):
    pass


# Create custom QTabWidget class
class QTabWidget(QW.QTabWidget, QWidget):
    """
    Defines the :class:`~QTabWidget` class.

    This class provides default settings and extra options for the
    :class:`~PyQt5.QtWidgets.QTabWidget` class.

    """

    # Signals
    currentIndexChanged = QC.Signal(int)
    currentTextChanged = QC.Signal(str)
    currentWidgetChanged = QC.Signal(QW.QWidget)
    tabTextChanged = QC.Signal(int, str)
    tabWasInserted = QC.Signal([int], [int, str], [int, QW.QWidget])
    tabWasRemoved = QC.Signal(int)

    # Override constructor to connect some signals
    def __init__(self, *args, browse_tabs=False, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Set up the tab widget
        self.init(browse_tabs)

    # This function sets up the tab widget
    def init(self, browse_tabs):
        # Set default tabbar
        self.setTabBar(QTabBar())

        # Connect signals
        self.currentChanged.connect(self.currentIndexChanged)
        self.currentChanged.connect(
            lambda index: self.currentTextChanged.emit(self.tabText(index)))
        self.currentChanged.connect(
            lambda index: self.currentWidgetChanged.emit(self.widget(index)))

        # Check if a browse menu was requested
        if browse_tabs:
            # Create a menu containing all available tabs
            browse_menu = QMenu('Browse', parent=self)
            browse_menu.aboutToShow.connect(self.update_browse_menu)
            self.browse_menu = browse_menu

            # Create a toolbutton for browsing all available tabs
            browse_but = QToolButton()
            browse_but.setText('V')
            browse_but.setToolTip("Browse tabs")
            browse_but.setMenu(browse_menu)
            browse_but.setPopupMode(browse_but.InstantPopup)
            self.browse_but = browse_but

            # Set browse button as the left corner widget
            self.setCornerWidget(browse_but, QC.Qt.TopLeftCorner)

    # This function updates the browse menu
    def update_browse_menu(self):
        """
        Updates the browse menu that shows all available tabs.

        """

        # Remove all actions currently in the browse menu
        self.browse_menu.clear()

        # Loop over all available tabs
        for i, name in enumerate(self.tabNames()):
            # Create a toggleable action for this tab
            tab_act = QAction(
                self, name,
                icon=self.tabIcon(i),
                tooltip=self.tabToolTip(i),
                toggled=lambda *args, index=i: self.setCurrentIndex(index))

            # If this tab is currently selected, check it
            tab_act.setChecked(self.currentIndex() == i)

            # Add action to menu
            self.browse_menu.addAction(tab_act)

    # Override addTab to automatically translate the given tab name
    def addTab(self, widget, label, icon=None):
        # Call super method
        if icon is None:
            return(super().addTab(widget, label))
        else:
            return(super().addTab(widget, icon, label))

    # Override setTabBar to automatically connect some signals
    def setTabBar(self, tabbar):
        # Connect the tabTextChanged signals
        tabbar.tabTextChanged.connect(self.tabTextChanged)

        # Call super method
        return(super().setTabBar(tabbar))

    # Override tabInserted to emit a signal whenever it is called
    def tabInserted(self, index):
        # Emit tabWasInserted signal
        self.tabWasInserted[int].emit(index)
        self.tabWasInserted[int, str].emit(index, self.tabText(index))
        self.tabWasInserted[int, QW.QWidget].emit(index, self.widget(index))

        # Call super method
        super().tabInserted(index)

    # Override tabRemoved to emit a signal whenever it is called
    def tabRemoved(self, index):
        # Emit tabWasRemoved signal
        self.tabWasRemoved.emit(index)

        # Call super method
        super().tabRemoved(index)

    # Define function that returns a list of all tab names
    def tabNames(self):
        return(list(map(self.tabText, range(self.count()))))

    # Define function that returns a list of all tab widgets
    def tabWidgets(self):
        return(list(map(self.widget, range(self.count()))))


# Create custom QTextEdit class
class QTextEdit(QW.QTextEdit, QWidget):
    pass


# Create custom QToolbar class
class QToolBar(QW.QToolBar, QWidget):
    """
    Defines the :class:`~QToolBar` class.

    This class provides default settings and extra options for the
    :class:`~PyQt5.QtWidgets.QToolBar` class.

    """

    def __init__(self, name, title=None, parent=None):
        # Save name
        self.name = name

        # If title is None, set it to name
        if title is None:
            title = name

        # Call super constructor
        super().__init__(title, parent)

    # This function retrieves the action of a menu and adds it to the toolbar
    def addMenu(self, menu):
        # Obtain the action associated with this menu
        action = menu.menuAction()

        # Add this action
        self.addAction(action)


# Create custom QToolButton class
class QToolButton(QW.QToolButton, QAbstractButton):
    # Override constructor to set some default settings
    def __init__(self, *args, **kwargs):
        # Call super constructor
        super().__init__(*args, **kwargs)

        # Use default settings
        self.setAutoRaise(True)


# Create custom QToolTip class
class QToolTip(QW.QToolTip):
    pass



# -*- coding: utf-8 -*-

"""
Figure Options
==============

"""


# %% IMPORTS
# Built-in imports

# Package imports
import matplotlib as mpl
from matplotlib import rcParams
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.plugins.figure.widgets.plot_entry import FigurePlotEntry
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['FigureOptionsDialog']


# %% CLASS DEFINITIONS
# Define class for the Figure options dialog
class FigureOptionsDialog(GW.QDialog):
    # Create applying and discarding signals
    applying = QC.Signal()
    discarding = QC.Signal()

    # Create refreshing_plots and refreshing_figure signals
    refreshing_plots = QC.Signal()
    refreshing_figure = QC.Signal()

    # Initialize FigureOptionsDialog
    def __init__(self, toolbar, *args, **kwargs):
        # Save provided FigureToolbar object
        self.toolbar = toolbar
        self.canvas = toolbar.canvas
        self.figure = self.canvas.figure
        self.axis = self.figure.gca()

        # Call super constructor
        super().__init__(toolbar)

        # Set up the figure options dialog
        self.init(*args, **kwargs)

    # This function sets up the figure options dialog
    def init(self):
        # Initialize empty dict of options
        self.options_dict = {}

        # Install event filter
        self.installEventFilter(self)

        # Set dialog properties
        # TODO: Figure out how to make the menu modal without it moving
        # automatically to the center on Linux
        # self.setWindowModality(QC.Qt.ApplicationModal)
        self.setWindowFlags(
            QC.Qt.Dialog |
            QC.Qt.FramelessWindowHint |
            QC.Qt.NoDropShadowWindowHint)

        # Create a layout
        layout = GL.QVBoxLayout(self)

        # Add the options tabs to it
        self.options_tabs = self.create_options_tabs()
        layout.addWidget(self.options_tabs)

        # Add stretch
        layout.addStretch()

        # Add a buttonbox
        button_box = QW.QDialogButtonBox()
        button_box.clicked.connect(self.buttonWasPressed)
        layout.addWidget(button_box)
        self.button_box = button_box

        # Add an 'Ok' button
        ok_but = button_box.addButton(button_box.Ok)
        ok_but.setToolTip("Apply current changes and close the figure options "
                          "menu")

        # Add a 'Cancel' button
        cancel_but = button_box.addButton(button_box.Cancel)
        cancel_but.setToolTip("Discard current changes and close the figure "
                              "options menu")
        self.cancel_but = cancel_but

        # Add an 'Apply' button
        apply_but = button_box.addButton(button_box.Apply)
        apply_but.setToolTip("Apply current figure options")
        self.apply_but = apply_but
        self.disable_apply_button()

        # Add a 'Refresh' button
        refresh_but = button_box.addButton('Refresh', button_box.ActionRole)
        refresh_but.setToolTip("Refresh figure using current figure options "
                               "without applying them")

        # Create a slot dict for all buttons
        self.slot_dict = {
            button_box.AcceptRole: [
                self.refresh_figure,
                self.apply_options,
                self.toolbar.toggle_options_dialog],
            button_box.RejectRole: [
                self.discard_options,
                self.refresh_figure,
                self.toolbar.toggle_options_dialog],
            button_box.ApplyRole: [
                self.apply_options],
            button_box.ActionRole: [
                self.refresh_figure]}

    # This function is called whenever a button is pressed
    @QC.Slot(QW.QAbstractButton)
    def buttonWasPressed(self, button):
        """
        Handles the actions that should be carried out when the provided
        `button` is pressed.

        """

        # Obtain the role of the provided button
        but_role = self.button_box.buttonRole(button)

        # Obtain the button's slots that must be called
        slots = self.slot_dict[but_role]

        # Call all slots defined in slots
        for slot in slots:
            slot()

    # This function adds the provided widget as an options entry
    def add_options_entry(self, widget):
        """
        Adds the provided `widget` as an options entry to this options dialog.
        This allows for the values of `widget` to be tracked and potentially
        discarded/reverted.

        """

        # Add widget as an entry to options_dict
        self.options_dict[widget] = get_box_value(widget)
        get_modified_signal(widget).connect(self.enable_apply_button)

    # This function removes the provided widget as an options entry
    def remove_options_entry(self, widget):
        """
        Removes the options entry associated with the provided `widget`, no
        longer tracking its values.

        """

        # Remove widget as an entry to options_dict
        self.options_dict.pop(widget)

    # This function creates the options tabwidget
    def create_options_tabs(self):
        # Create a tab widget
        tab_widget = GW.QTabWidget()

        # Add figure tab
        tab_widget.addTab(*self.create_figure_tab())

        # Add plots tab
        tab_widget.addTab(*self.create_plots_tab())

        # Return layout
        return(tab_widget)

    # This function creates the 'Figure' tab
    def create_figure_tab(self):
        # Create a tab
        tab = GW.BaseBox()
        get_modified_signal(tab).connect(self.enable_apply_button)

        # Create layout
        layout = GL.QFormLayout(tab)

        # Make line edit for title
        title_box = GW.FigureLabelBox()
        title_box[0].setToolTip("Figure title")
        title_box[1].setToolTip("Title size")
        set_box_value(
            title_box, ('', {'fontsize': rcParams['figure.titlesize']}))
        self.add_options_entry(title_box)
        self.refreshing_figure.connect(lambda: self.axis.set_title(
            *get_box_value(title_box, str, dict)))
        layout.addRow("Title", title_box)
        self.title_box = title_box

        # X-AXIS
        # Create a group box for the X-axis
        x_axis_group = GW.QGroupBox("X-axis")
        layout.addRow(x_axis_group)
        x_axis_layout = GL.QFormLayout(x_axis_group)

        # Make a box for setting the label on the x-axis
        x_label_box = GW.FigureLabelBox()
        x_label_box[0].setToolTip("Label of the X-axis")
        x_label_box[1].setToolTip("Label size")
        set_box_value(
            x_label_box, ('', {'fontsize': rcParams['axes.labelsize']}))
        self.add_options_entry(x_label_box)
        self.refreshing_figure.connect(lambda: self.axis.set_xlabel(
            *get_box_value(x_label_box, str, dict)))
        x_axis_layout.addRow("Label", x_label_box)
        self.x_label_box = x_label_box

        # Make a box for setting the range on the x-axis
        # TODO: Allow for a small formula or cell range to be given instead?
        x_range_box = GW.DualLineEdit((float, float),
                                      r"<html>&le; X &le;</html>")
        x_min_box, x_max_box = x_range_box[:]
        x_min_box.setToolTip("Minimum value of the X-axis")
        x_max_box.setToolTip("Maximum value of the X-axis")
        set_box_value(x_range_box, self.axis.get_xlim())

        # Connect signals for x_range_box
        self.refreshing_figure.connect(lambda: self.axis.set_xlim(
            *get_box_value(x_range_box), auto=None))
        self.axis.callbacks.connect(
            'xlim_changed', lambda x: set_box_value(x_range_box, x.get_xlim()))

        # Make togglebox for enabling/disabling the use of this range
        x_range_togglebox = GW.ToggleBox(
            x_range_box, tooltip="Check to manually set the X-axis range")
        self.add_options_entry(x_range_togglebox)
        x_axis_layout.addRow("Range", x_range_togglebox)

        # Connect signals for x_range_togglebox
        self.refreshing_figure.connect(lambda: self.axis.set_autoscalex_on(
            not get_box_value(x_range_togglebox, bool)))

        # Make a box for setting the scale on the x-axis
        x_scale_box = GW.QComboBox()
        x_scale_box.addItems(['linear', 'log', 'symlog', 'logit'])
        x_scale_box.setToolTip("Value scale of the X-axis")
        self.add_options_entry(x_scale_box)
        self.refreshing_figure.connect(lambda: self.axis.set_xscale(
            get_box_value(x_scale_box)))
        x_axis_layout.addRow("Scale", x_scale_box)

        # Y-AXIS
        # Create a group box for the Y-axis
        y_axis_group = GW.QGroupBox("Y-axis")
        layout.addRow(y_axis_group)
        y_axis_layout = GL.QFormLayout(y_axis_group)

        # Make a box for setting the label on the y-axis
        y_label_box = GW.FigureLabelBox()
        y_label_box[0].setToolTip("Label of the Y-axis")
        y_label_box[1].setToolTip("Label size")
        set_box_value(
            y_label_box, ('', {'fontsize': rcParams['axes.labelsize']}))
        self.add_options_entry(y_label_box)
        self.refreshing_figure.connect(lambda: self.axis.set_ylabel(
            *get_box_value(y_label_box, str, dict)))
        y_axis_layout.addRow("Label", y_label_box)
        self.y_label_box = y_label_box

        # Make a box for setting the range on the y-axis
        y_range_box = GW.DualLineEdit((float, float),
                                      r"<html>&le; Y &le;</html>")
        y_min_box, y_max_box = y_range_box[:]
        y_min_box.setToolTip("Minimum value of the Y-axis")
        y_max_box.setToolTip("Maximum value of the Y-axis")
        set_box_value(y_range_box, self.axis.get_ylim())

        # Connect signals for y_range_box
        self.refreshing_figure.connect(lambda: self.axis.set_ylim(
            *get_box_value(y_range_box), auto=None))
        self.axis.callbacks.connect(
            'ylim_changed', lambda y: set_box_value(y_range_box, y.get_ylim()))

        # Make togglebox for enabling/disabling the use of this range
        y_range_togglebox = GW.ToggleBox(
            y_range_box, tooltip="Check to manually set the Y-axis range")
        self.add_options_entry(y_range_togglebox)
        y_axis_layout.addRow("Range", y_range_togglebox)

        # Connect signals for y_range_togglebox
        self.refreshing_figure.connect(lambda: self.axis.set_autoscaley_on(
            not get_box_value(y_range_togglebox, bool)))

        # Make a box for setting the scale on the y-axis
        y_scale_box = GW.QComboBox()
        y_scale_box.addItems(['linear', 'log', 'symlog', 'logit'])
        y_scale_box.setToolTip("Value scale of the Y-axis")
        self.add_options_entry(y_scale_box)
        self.refreshing_figure.connect(lambda: self.axis.set_yscale(
            get_box_value(y_scale_box)))
        y_axis_layout.addRow("Scale", y_scale_box)

        # PROPS
        # Create a group box for figure properties
        props_group = GW.QGroupBox("Properties")
        layout.addRow(props_group)
        props_layout = GL.QFormLayout(props_group)

        # Make a combobox for choosing the location of the legend
        legend_loc_box = GW.QComboBox()
        legend_loc_box.addItems(mpl.legend.Legend.codes.keys())
        legend_loc_box.setToolTip("Location of the figure legend")
        set_box_value(legend_loc_box, rcParams['legend.loc'])

        # Make a togglebox for using a legend
        legend_togglebox = GW.ToggleBox(
            legend_loc_box, "Legend",
            tooltip="Toggle the use of a figure legend")
        self.add_options_entry(legend_togglebox)
        props_layout.addRow(legend_togglebox)
        self.legend_togglebox = legend_togglebox

        # Return tab
        return(tab, "Figure")

    # This function creates the 'Plots' tab
    def create_plots_tab(self):
        # Create a tab
        tab = GW.BaseBox()
        get_modified_signal(tab).connect(self.enable_apply_button)

        # Create layout
        layout = GL.QFormLayout(tab)

        # PLOT
        # Create a plot picker layout
        plot_layout = GL.QHBoxLayout()
        layout.addRow(plot_layout)

        # Create a label
        plot_label = GW.QLabel("Plot")
        plot_label.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
        plot_layout.addWidget(plot_label)

        # Create a combobox for choosing an existing plot
        plot_entries = GW.QComboBox()
        plot_entries.setToolTip("Select the plot entry you wish to edit")
        plot_layout.addWidget(plot_entries)
        get_modified_signal(plot_entries).disconnect(tab.modified)
        self.plot_entries = plot_entries

        # Add a toolbutton for adding a new plot entry
        add_but = GW.QToolButton()
        add_but.setToolTip("Add new plot entry")
        get_modified_signal(add_but).connect(self.add_entry)
        plot_layout.addWidget(add_but)

        # If this theme has an 'add' icon, use it
        if QG.QIcon.hasThemeIcon('add'):
            add_but.setIcon(QG.QIcon.fromTheme('add'))
        # Else, use a simple plus
        else:
            add_but.setText('+')

        # Add a separator
        layout.addSeparator()

        # Add a stacked widget here for dividing the plots
        plot_pages = GW.QStackedWidget()
        get_modified_signal(plot_entries, int).connect(
            plot_pages.setCurrentIndex)
        layout.addRow(plot_pages)
        self.plot_pages = plot_pages

        # Return tab
        return(tab, "Plots")

    # This function adds a plot entry
    @QC.Slot()
    def add_entry(self):
        # Obtain index of this plot entry
        index = self.plot_entries.count()

        # Obtain name
        name = "%i_plot" % (index)

        # Create plot entry box
        plot_entry = FigurePlotEntry(index, name, self.toolbar)

        # Connect signals
        plot_entry.entryNameChanged.connect(
            lambda x: self.plot_entries.setItemText(index, x))
        plot_entry.entryRemoveRequested.connect(self.remove_entry)

        # Add it to the plot_entries and plot_pages
        self.plot_entries.addItem(name)
        self.plot_pages.addWidget(plot_entry)
        get_modified_signal(plot_entry).connect(self.enable_apply_button)

        # Set the shown entry to the new entry
        set_box_value(self.plot_entries, index)

    # This function removes a plot entry
    @QC.Slot()
    def remove_entry(self):
        # Obtain the index and widget of the currently shown plot entry
        index = get_box_value(self.plot_entries, int)
        name = get_box_value(self.plot_entries, str)
        widget = self.plot_pages.currentWidget()

        # If index is -1, return
        if(index == -1):
            return

        # Show a warning message asking if the user really wants to remove it
        # TODO: Should this be removed once changes can be discarded?
        button_clicked = GW.QMessageBox.warning(
            self, "WARNING: Delete plot",
            ("Are you sure you want to delete the plot with name <b>%s</b>? "
             "(<i>Note: This action is irreversible!</i>)" % (name)),
            GW.QMessageBox.Yes | GW.QMessageBox.No, GW.QMessageBox.No)

        # Remove the entry and page at this index if the user answered 'yes'
        if(button_clicked == GW.QMessageBox.Yes):
            self.plot_entries.removeItem(index)
            self.plot_pages.removeWidget(widget)
            widget.setParent(None)
            widget.close()
            del widget

    # Override showEvent to show the dialog in the proper location
    def showEvent(self, event):
        # Call super event
        super().showEvent(event)

        # Determine the position of the top left corner of the figure dock
        dock_pos = self.toolbar.rect().topLeft()

        # Determine the size of this dialog
        size = self.size()
        size = QC.QPoint(size.width(), 0)

        # Determine position of top left corner
        dialog_pos = self.toolbar.mapToGlobal(dock_pos-size)

        # Move it slightly to give some spacing
        dialog_pos.setX(dialog_pos.x()-12)

        # Move the dialog there
        self.move(dialog_pos)

    # Override eventFilter to filter out ESC presses
    def eventFilter(self, widget, event):
        # Check if the event involves anything for which the popup should close
        if((event.type() == QC.QEvent.KeyPress) and
           event.key() in (QC.Qt.Key_Escape,)):
            # Act as if the 'Cancel' button was clicked
            self.cancel_but.click()
            return(True)

        # Else, process events as normal
        else:
            return(super().eventFilter(widget, event))

    # This function refreshes the figure
    # OPTIMIZE: Drawing canvas can take several seconds for large data plots
    @QC.Slot()
    def refresh_figure(self):
        # Emit the refreshing signals
        self.refreshing_plots.emit()
        self.refreshing_figure.emit()

        # Update the legend
        self.set_legend()

        # Process figure axes limits
        self.axis.relim()
        self.axis.autoscale_view(None, True, True)

        # Draw canvas
        self.canvas.draw()

    # This function sets the legend of the figure
    @QC.Slot()
    def set_legend(self):
        # Obtain the legend_flag
        flag, loc = get_box_value(self.legend_togglebox)

        # If flag is True, create a legend
        if flag:
            self.axis.legend(loc=loc)

        # Else, remove the current one if it exists
        elif self.axis.legend_ is not None:
            self.axis.legend_.remove()

    # This function applies the new figure options values
    @QC.Slot()
    def apply_options(self):
        """
        Applies all current values of all figure options.

        """

        # Emit the applying signal
        self.applying.emit()

        # Apply all new values
        for widget in self.options_dict.keys():
            self.options_dict[widget] = get_box_value(widget)

        # Disable the apply button
        self.disable_apply_button()

    # This function enables the apply button
    @QC.Slot()
    def enable_apply_button(self):
        """
        Qt slot that enables the apply button at the bottom of the figure
        options dialog.
        The apply button is enabled if at least one change has been made to any
        figure option.

        """

        self.apply_but.setEnabled(True)

    # This function disables the apply button
    @QC.Slot()
    def disable_apply_button(self):
        """
        Qt slot that disables the apply button at the bottom of the figure
        options dialog.
        The apply button is disabled whenever no changes have been made to any
        figure option.

        """

        self.apply_but.setEnabled(False)

    # This function discards all changes to the figure options
    @QC.Slot()
    def discard_options(self):
        """
        Discards the current values of all figure options and sets them back to
        their saved values.

        """

        # Emit the discarding signal
        self.discarding.emit()

        # Discard all current changes
        for widget, value in self.options_dict.items():
            set_box_value(widget, value)

        # Disable the apply button
        self.disable_apply_button()



# -*- coding: utf-8 -*-

"""
Data Table Widget
=================

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.plugins.data_table.widgets.view import DataTableView
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['DataTableWidget']


# %% CLASS DEFINITIONS
# Define class for the DataTable widget
class DataTableWidget(GW.QWidget):
    # Initialize DataTableWidget
    def __init__(self, parent=None, *args, **kwargs):
        # Call super constructor
        super().__init__(parent)

        # Set up the data table widget
        self.init(*args, **kwargs)

    # This function sets up the data table widget
    def init(self, import_func=None):
        # Create a layout
        layout = GL.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        # Create a dimensions layout
        dimensions_layout = GL.QHBoxLayout()
        layout.addLayout(dimensions_layout)

        # Add a label to this layout
        dimensions_layout.addWidget(GW.QLabel('Dimensions: '))

        # Create dual spinbox for setting n_rows and n_cols
        dimensions_box = GW.DualSpinBox((int, int), "X")
        dimensions_layout.addWidget(dimensions_box)
        n_rows_box, n_cols_box = dimensions_box[:]
        n_rows_box.setRange(0, 9999999)
        n_rows_box.setToolTip("Number of rows in this data table (max. %i)"
                              % (n_rows_box.maximum()))
        n_cols_box.setRange(0, 702)
        n_cols_box.setToolTip("Number of columns in this data table (max. %i)"
                              % (n_cols_box.maximum()))
        self.dimensions_box = dimensions_box

        # Create a layout for applying or reverting the dimensions
        buttons_layout = GL.QHBoxLayout()
        buttons_layout.setContentsMargins(0, 0, 0, 0)
        buttons_layout.setSpacing(0)
        dimensions_layout.addLayout(buttons_layout)

        # If this theme has a 'cancel' icon, use it
        if QG.QIcon.hasThemeIcon('cancel'):
            rev_icon = QG.QIcon.fromTheme('cancel')
        # Else, use a standard icon
        else:
            rev_icon = self.style().standardIcon(
                QW.QStyle.SP_DialogCancelButton)

        # Create a revert toolbutton
        rev_but = GW.QToolButton()
        rev_but.setToolTip("Revert to current data table dimensions")
        rev_but.setIcon(rev_icon)
        get_modified_signal(rev_but).connect(self.revert_table_dimensions)
        buttons_layout.addWidget(rev_but)

        # If this theme has an 'apply' icon, use it
        if QG.QIcon.hasThemeIcon('apply'):
            app_icon = QG.QIcon.fromTheme('apply')
        # Else, use a standard icon
        else:
            app_icon = self.style().standardIcon(
                QW.QStyle.SP_DialogApplyButton)

        # Create an apply toolbutton
        app_but = GW.QToolButton()
        app_but.setToolTip("Apply new data table dimensions")
        app_but.setIcon(app_icon)
        get_modified_signal(app_but).connect(self.apply_table_dimensions)
        buttons_layout.addWidget(app_but)

        # Add a stretcher
        dimensions_layout.addStretch()

        # Create the DataTableView object
        self.view = DataTableView(self, import_func)

        # Set initial values of the spinboxes
        self.revert_table_dimensions()

        # Connect signals from data table view
        self.view.model().rowCountChanged.connect(
            lambda x: set_box_value(n_rows_box, x))
        self.view.model().columnCountChanged.connect(
            lambda x: set_box_value(n_cols_box, x))
        self.view.model().lastColumnRemoved.connect(
            lambda: n_rows_box.setEnabled(False))
        self.view.model().firstColumnInserted.connect(
            lambda: n_rows_box.setEnabled(True))

        # Add data_table to the layout
        layout.addWidget(self.view)

    # Override closeEvent to perform some additional clean-up
    def closeEvent(self, *args, **kwargs):
        # Close the data table view
        self.view.close()

        # Call super event
        super().closeEvent(*args, **kwargs)

    # This property returns the model of the DataTableView in this data table
    @property
    def model(self):
        return(self.view.model())

    # This function applies the table dimensions as requested by the user
    @QC.Slot()
    def apply_table_dimensions(self):
        # Obtain the values of the dimensions_box
        n_rows, n_cols = get_box_value(self.dimensions_box)

        # Set the rows and columns in the model
        self.view.setRowCount(n_rows)
        self.view.setColumnCount(n_cols)

    # This function reverts the table dimensions back to their current values
    @QC.Slot()
    def revert_table_dimensions(self):
        # Obtain the current table dimensions
        n_rows = self.view.rowCount()
        n_cols = self.view.columnCount()

        # Set the spinboxes to the proper values
        set_box_value(self.dimensions_box, (n_rows, n_cols))



# -*- coding: utf-8 -*-

"""
Histogram Type
==============

"""


# %% IMPORTS
# Built-in imports

# Package imports
from matplotlib import rcParams
import numpy as np
import pandas as pd
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy.plugins.figure.widgets.types import BasePlotType
from guipy.widgets import get_box_value, set_box_value

# All declaration
__all__ = ['HistogramType']


# %% CLASS DEFINITIONS
# Create custom class for making a histogram plot
class HistogramType(BasePlotType):
    """
    Provides the definition of the :class:`~HistogramType` plot type.

    """

    # Class attributes
    NAME = "Histogram"
    PREFIX = "hist"
    AXIS_TYPE = "2D"
    PROP_NAMES = [*BasePlotType.PROP_NAMES, 'MultiHistData', 'Histogram']

    # This function sets up the histogram plot
    def init(self, *args, **kwargs):
        # Create layout for this histogram plot
        super().init(*args, **kwargs)

        # Manually call hist_tab_added
        self.hist_tab_added(0)

        # Set current hist_kwargs to None
        self.hist_kwargs = None

    # This function is called whenever a new histogram data tab is added
    @QC.Slot(int)
    def hist_tab_added(self, index):
        """
        This slot is automatically called whenever a new histogram data tab
        with `index` has been added, and is used for initializing certain
        aspects of this tab.

        """

        # Obtain the number of data tabs already present
        n_tabs = self.multi_data_box.count()

        # Use this number to cycle through MPL's color cycler
        color = "C%i" % (n_tabs-1 % len(rcParams['axes.prop_cycle']))
        set_box_value(self.multi_data_box, color, index, 'hist_color_box')

    # This function draws the histogram plot
    # TODO: Allow for the bin-width and a single bin-edge to be given instead?
    @QC.Slot()
    def draw_plot(self):
        # Obtain the x columns
        try:
            xcols = get_box_value(self.multi_data_box, 'x_data_box', 1)
        # If any column cannot be called, return
        except IndexError:
            self.remove_plot()
            return

        # If any of the xcols contains None, return as well
        if pd.isna(xcols).any():
            self.remove_plot()
            return

#        # Determine if a custom value range is requested
#        value_range = get_box_value(self.hist_range_box)
#        value_range = value_range[1:] if value_range[0] else None

        # Obtain the histogram keyword arguments
        bins = get_box_value(self.n_bins_box)
        hist_kwargs = {
            'bins': bins if bins else self.n_bins_box.specialValueText(),
            'cumulative': get_box_value(self.hist_cumul_box),
            'orientation': get_box_value(self.hist_orient_box).lower()}

        # Check if this histogram should be drawn
        if self.plot is not None:
            # Check all arguments except data
            if not all(map(lambda k: (hist_kwargs[k] == self.hist_kwargs[k]),
                           hist_kwargs.keys())):
                pass

            # If all are the same, check data
            elif((len(xcols) == len(self.hist_kwargs['xcols'])) and
                 all(map(np.array_equal, xcols, self.hist_kwargs['xcols']))):
                return

        # As histograms cannot be modified, remove current one
        self.remove_plot()

        # Make a new histogram
        self.axis.hist(xcols, **hist_kwargs)

        # Save the made histogram(s)
        self.plot = self.axis.containers[-len(xcols):]

        # Loop over all labels for all data sets
        for i, label in enumerate(get_box_value(self.multi_data_box,
                                                'data_label_box')):
            # If label is not empty, reuse it in the plot
            if label:
                self.plot[i].set_label(label)
            # Else, obtain its label from MPL
            else:
                label = self.plot[i].get_label()
                set_box_value(self.multi_data_box, label, i, 'data_label_box')

        # Save what arguments were used for this histogram
        hist_kwargs['xcols'] = xcols
        self.hist_kwargs = hist_kwargs

    # This function updates the histogram plot
    @QC.Slot()
    def update_plot(self):
        # Draw the plot
        self.draw_plot()

        # If histograms currently exist, update them
        if self.plot is not None:
            # Obtain the labels and colors of all elements
            labels = get_box_value(self.multi_data_box, 'data_label_box')
            colors = get_box_value(self.multi_data_box, 'hist_color_box')

            for i, (plot, label, color) in enumerate(zip(self.plot, labels,
                                                         colors)):
                # Set label
                plot.set_label(label)

                # Update bin colors
                for patch in plot.patches:
                    patch.set_color(color)

    # This function removes the histogram from the figure
    @QC.Slot()
    def remove_plot(self):
        """
        Removes all histograms handled by this instance from the current
        figure.

        """

        # Remove the plots from the figure if they exist
        if self.plot is not None:
            for plot in self.plot:
                # Remove all patches of this histogram
                for patch in plot.patches:
                    self.axis.patches.remove(patch)

                # And remove the container for it as well
                self.axis.containers.remove(plot)

            # Set plot and hist_kwargs to None
            self.plot = None
            self.hist_kwargs = None



# -*- coding: utf-8 -*-

"""
Scatter Type
============

"""


# %% IMPORTS
# Built-in imports

# Package imports
from matplotlib import rcParams
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy.plugins.figure.widgets.types import BasePlotType
from guipy.widgets import get_box_value, set_box_value

# All declaration
__all__ = ['ScatterType']


# %% CLASS DEFINITIONS
# Create custom class for making a scatter plot
# TODO: Figure out how to manipulate plt.scatter instead of plt.plot
class ScatterType(BasePlotType):
    """
    Provides the definition of the :class:`~ScatterType` plot type.

    """

    # Class attributes
    NAME = "Scatter"
    PREFIX = "scatter"
    AXIS_TYPE = "2D"
    PROP_NAMES = [*BasePlotType.PROP_NAMES, 'Data2D', 'ScatterMarker']

    # This function sets up the scatter plot
    def init(self, *args, **kwargs):
        # Create layout for this scatter plot
        super().init(*args, **kwargs)

        # Set the starting color to be the number of scatters already present
        n_lines = len(self.axis.lines)
        color = "C%i" % (n_lines % len(rcParams['axes.prop_cycle']))
        set_box_value(self.marker_color_box, color)

    # This function draws the 2D scatter plot
    @QC.Slot()
    def draw_plot(self):
        # Obtain the x and y columns
        try:
            xcol = get_box_value(self.x_data_box)[1]
            ycol = get_box_value(self.y_data_box)[1]
        # If any of the columns cannot be called, return
        except IndexError:
            self.remove_plot()
            return

        # If either xcol or ycol is None, return
        if xcol is None or ycol is None:
            self.remove_plot()
            return

        # If xcol and ycol are not the same shape, return
        if(len(xcol) != len(ycol)):
            self.remove_plot()
            return

        # If the current saved scatter is not already in the figure, make one
        if self.plot not in self.axis.lines:
            # Make and update plot
            self.plot = self.axis.plot(xcol, ycol)[0]
            self.plot.set_linestyle('')

            # Obtain label currently set in label box
            label = get_box_value(self.data_label_box)

            # If label is not empty, reuse it in the plot
            if label:
                self.plot.set_label(label)
            # Else, obtain its label from MPL
            else:
                set_box_value(self.data_label_box, self.plot.get_label())

            # If the figure currently has no title, set it
            title_box = self.toolbar.options_dialog.title_box[0]
            if not get_box_value(title_box):
                set_box_value(title_box, "%s vs. %s" % (xcol.name, ycol.name))

            # If the figure currently has no axes labels, set them
            x_label_box = self.toolbar.options_dialog.x_label_box[0]
            y_label_box = self.toolbar.options_dialog.y_label_box[0]
            if not (get_box_value(x_label_box) or get_box_value(y_label_box)):
                set_box_value(x_label_box, xcol.name)
                set_box_value(y_label_box, ycol.name)

        # If it does exist, check if it requires updating
        else:
            # Obtain the data currently used for this plot
            xcol_cur = self.plot.get_xdata()
            ycol_cur = self.plot.get_ydata()

            # If there are differences, update plot
            if not (xcol_cur == xcol).all():
                self.plot.set_xdata(xcol)
            if not (ycol_cur == ycol).all():
                self.plot.set_ydata(ycol)

    # This function updates the 2D scatter plot
    @QC.Slot()
    def update_plot(self):
        # Draw the plot
        self.draw_plot()

        # If scatter currently exists, update it
        if self.plot is not None:
            # Set label
            self.plot.set_label(get_box_value(self.data_label_box))

            # Update marker style, size and color
            self.plot.set_marker(get_box_value(self.marker_style_box))
            self.plot.set_markersize(get_box_value(self.marker_size_box))
            self.plot.set_markeredgecolor(get_box_value(self.marker_color_box))
            self.plot.set_markerfacecolor(get_box_value(self.marker_color_box))

    # This function removes the 2D scatter plot
    @QC.Slot()
    def remove_plot(self):
        # Remove the plot from the figure if it exists
        if self.plot in self.axis.lines:
            self.axis.lines.remove(self.plot)

            # Set plot to None
            self.plot = None



# -*- coding: utf-8 -*-

"""
Checkboxes
==========

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['ToggleBox']


# %% CLASS DEFINITIONS
# Make class that toggles a widget
class ToggleBox(GW.DualBaseBox):
    """
    Defines the :class:`~ToggleBox` class.

    This widget allows for a provided widget to be toggled on and off.

    """

    # Signals
    modified = QC.Signal([], [bool])

    # Initialize the ToggleBox class
    def __init__(self, widget, text=None, tooltip=None, parent=None):
        """
        Initialize an instance of the :class:`~ToggleBox` class.

        Parameters
        ----------
        widget : :obj:`~PyQt5.QtWidgets.QWidget` object
            The widget that must be made toggleable.

        Optional
        --------
        text : str or None. Default: None
            The text that must be used for the checkbox.
            If *None*, the checkbox has no text.
        tooltip : str or None. Default: None
            The text that must be set as the tooltip of the checkbox.
            If *None*, the checkbox has no tooltip.
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this togglebox or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the togglebox
        self.init(widget, text, tooltip)

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[bool])

    # This function sets up the togglebox
    def init(self, widget, text, tooltip):
        """
        Sets up the togglebox after it has been initialized.

        """

        # Create the box_layout
        box_layout = GL.QHBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create a checkbox for toggling the widget
        if text is not None:
            checkbox = GW.QCheckBox(text)
        else:
            checkbox = GW.QCheckBox()
        box_layout.addWidget(checkbox)
        self.checkbox = checkbox
        self.left_box = checkbox

        # Set tooltip
        if tooltip is not None:
            checkbox.setToolTip(tooltip)

        # Add the widget to it
        box_layout.addWidget(widget)
        self.widget = widget
        self.right_box = widget

        # Connect some signals
        get_modified_signal(checkbox).connect(widget.setEnabled)

        # Make sure that the checkbox is unchecked
        set_box_value(checkbox, False)
        widget.setEnabled(False)

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        # Emit modified signal with proper types
        self.modified[bool].emit(get_box_value(self.checkbox))

    # This function retrieves a value of this special box
    def get_box_value(self, *value_sig):
        """
        Returns the current values of this togglebox as a tuple.

        Returns
        -------
        value : bool or tuple
            A tuple containing the values of the checkbox and widget, formatted
            as `(checkbox, widget)`.
            If `value_sig` contains type 'bool', only the value of `checkbox`
            is returned.

        """

        # If solely the value of the checkbox was requested, return it
        if bool in value_sig:
            return(get_box_value(self.checkbox))
        # Else, return the checkbox and widget values
        else:
            return(get_box_value(self.checkbox),
                   get_box_value(self.widget, *value_sig))

    # This function sets the value of this special box
    def set_box_value(self, value, *value_sig):
        """
        Sets the current value of the togglebox to `value`.

        Parameters
        ----------
        value : tuple
            A tuple containing the values of the checkbox and widget, formatted
            as `(checkbox, widget)`.

        """

        set_box_value(self.checkbox, value[0])
        set_box_value(self.widget, value[1], *value_sig)



# -*- coding: utf-8 -*-

"""
CSV Formatter
=============

"""


# %% IMPORTS
# Package imports
import numpy as np
import pandas as pd

# GuiPy imports
from guipy.plugins.data_table.formatters import BaseFormatter

# All declaration
__all__ = ['CSVFormatter']


# %% CLASS DEFINITIONS
# Define Formatter for .csv-files
class CSVFormatter(BaseFormatter):
    # Class attributes
    TYPE = "Comma-Separated Values"
    EXTS = ['.csv']

    # Define the export to csv function
    def exporter(self, data_table, filepath):
        # Obtain the data in the data table
        data = data_table.model._data

        # Export it as a CSV-file
        data.to_csv(filepath, index=False)

    # Define the import from csv function
    def importer(self, filepath, parent=None):
        # Read in the first 2 lines of the CSV-file twice
        df_header = pd.read_csv(filepath, skipinitialspace=True, nrows=2)
        df_no_header = pd.read_csv(filepath, skipinitialspace=True, nrows=2,
                                   header=None)

        # Read in the CSV-file as a data frame depending on if it has a header
        if np.all(df_header.dtypes.values == df_no_header.dtypes.values):
            # If corresponding columns share dtypes, then it has no header
            data_frame = pd.read_csv(filepath, skipinitialspace=True,
                                     header=None)
        else:
            # Else, it must have a header
            data_frame = pd.read_csv(filepath, skipinitialspace=True)

        # Return data_frame
        return(data_frame)



# -*- coding: utf-8 -*-

"""
App Start
=========
Starts the *GuiPy* application.

"""


# %% IMPORTS
# Built-in imports
import signal

# Package imports
import matplotlib.pyplot as plt
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import APP_NAME
from guipy.app.main_window import MainWindow

# All declaration
__all__ = ['main']


# %% MAIN FUNCTION
def main():
    """
    Prepares an :obj:`~PyQt5.QtWidgets.QApplication` instance and starts the
    *GuiPy* application.

    """

    # Obtain application instance
    qapp = QW.QApplication.instance()

    # If qapp is None, create a new one
    if qapp is None:
        QW.QApplication.setAttribute(QC.Qt.AA_EnableHighDpiScaling)
        qapp = QW.QApplication([APP_NAME])

    # Set name of application
    qapp.setApplicationName(APP_NAME)

    # Make sure that the application quits when last window closes
    qapp.lastWindowClosed.connect(qapp.quit, QC.Qt.QueuedConnection)

    # Initialize main window and draw (show) it
    main_window = MainWindow()
    main_window.show()
    main_window.raise_()
    main_window.activateWindow()

    # Replace KeyboardInterrupt error by system's default handler
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Set MPL's backend to 'Agg'
    plt.switch_backend('Agg')

    # Start application
    qapp.exec_()


# %% MAIN EXECUTION
if(__name__ == '__main__'):
    main()



# -*- coding utf-8 -*-

"""
Plugins
=======
Contains the various different plugins that are used in *GuiPy*.

"""


# %% IMPORTS
# Import core modules
from . import base
from .base import *

# Import subpackages
from . import data_table, figure
from .data_table.plugin import *
from .figure.plugin import *

# All declaration
__all__ = ['base', 'data_table', 'figure']
__all__.extend(base.__all__)
__all__.extend(data_table.plugin.__all__)
__all__.extend(figure.plugin.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Colors
======
Provides a collection of :class:`~PyQt5.QtWidgets.QWidget` subclasses for
handling colors in :mod:`~matplotlib`.

"""


# %% IMPORTS
# Built-in imports
from importlib import import_module
from itertools import chain
import re

# Package imports
from cmasher.utils import get_cmap_type
from matplotlib import cm, rcParams
from matplotlib.colors import (
    BASE_COLORS, CSS4_COLORS, Colormap, to_hex, to_rgba)
import matplotlib.pyplot as plt
import numpy as np
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW
from sortedcontainers import SortedDict as sdict, SortedSet as sset

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.widgets import get_box_value, get_modified_signal, set_box_value

# All declaration
__all__ = ['ColorBox', 'ColorMapBox']


# %% CLASS DEFINITIONS
# Make class with a special box for setting the color of a plotted line
class ColorBox(GW.BaseBox):
    """
    Defines the :class:`~ColorBox` class.

    This widget allows for colors in *matplotlib* to be easily picked by the
    user.
    All cyclic; basic; and CSS4 colors in *matplotlib* are available from the
    combobox.
    Additionally, the HEX code of a color can be typed in as well.
    Finally, the user has access to a colorwheel by clicking on the color
    label, which allows for any color to be picked manually, including any
    colors currently on screen using a screenpicker.

    The default color, which is used when an invalid color is currently typed
    in, can be set using :meth:`~set_default_color`.

    """

    # Signals
    modified = QC.Signal([], [str])

    # Set the size for the color labels
    clabel_size = (70, 18)

    def __init__(self, add_cycler=True, parent=None):
        """
        Initialize an instance of the :class:`~ColorBox` class.

        Optional
        --------
        add_cycler : bool. Default: True
            Whether or not to add *matplotlib*'s cyclic colors to the list of
            preset colors.
        parent : :obj:`~PyQt5.QtCore.QObject` object or None. Default: None
            The parent object to use for this colorbox or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the color box
        self.init(add_cycler)

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[str])

    # This function creates the color box
    def init(self, add_cycler):
        """
        Sets up the color box entry after it has been initialized.

        This function is mainly responsible for creating the color wheel and
        color label, that allow the user to quickly cycle through different
        color options.

        """

        # Create the box layout
        box_layout = GL.QHBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create a color label
        color_label = self.create_color_label()
        self.color_label = color_label
        box_layout.addWidget(color_label)

        # Create a color combobox
        color_combobox = self.create_color_combobox(add_cycler)
        box_layout.addWidget(color_combobox)
        self.color_combobox = color_combobox

        # Set the default starting color of the color box
        self.set_box_value('C0')

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        self.modified[str].emit(get_box_value(self.color_combobox))

    # This function creates the color label
    def create_color_label(self):
        """
        Creates a special label that shows the currently selected or hovered
        color, and returns it.

        """

        # Create the color label
        color_label = GW.QLabel()

        # Set some properties
        color_label.setFrameShape(QW.QFrame.StyledPanel)
        color_label.setScaledContents(True)
        color_label.setToolTip("Click to open the custom color picker")
        color_label.setSizePolicy(QW.QSizePolicy.Fixed, QW.QSizePolicy.Fixed)
        color_label.mousePressed.connect(self.show_colorpicker)

        # Return it
        return(color_label)

    # This function creates the color combobox
    def create_color_combobox(self, add_cycler):
        """
        Creates a combobox that holds all default colors accepted by matplotlib
        and returns it.

        """

        # Obtain the CN colors if requested
        if add_cycler:
            n_cyclic = len(rcParams['axes.prop_cycle'])
            CN_COLORS = [("C%i" % (i), "This is MPL cyclic color #%i" % (i))
                         for i in range(n_cyclic)]
        else:
            CN_COLORS = []

        # Make tuple of all colors
        colors = (CN_COLORS, BASE_COLORS, CSS4_COLORS)

        # Determine the cumulative lengths of all four sets
        cum_len = np.cumsum(list(map(len, colors)))

        # Make combobox for colors
        color_box = GW.EditableComboBox()

        # Add a special validator to this combobox
        validator = GW.ComboBoxValidator(color_box, r"#?[\da-fA-F]{6}")
        color_box.setValidator(validator)

        # Fill combobox with all colors
        for i, color in enumerate(chain(*colors)):
            # If color is a tuple, it consists of (color, tooltip)
            if isinstance(color, tuple):
                color_box.addItem(color[0])
                color_box.setItemData(i, color[1], QC.Qt.ToolTipRole)
            else:
                color_box.addItem(color)

        # Add some separators
        for i in reversed(cum_len[:-1]):
            color_box.insertSeparator(i)

        # Set remaining properties
        color_box.setToolTip("Select or type (in HEX) the color")
        color_box.highlighted[str].connect(self.set_color_label)
        color_box.popup_hidden[str].connect(self.set_color)
        color_box.editTextChanged.connect(self.set_color)
        color_box.focusLost.connect(
            lambda: self.set_color(self.get_box_value()))
        return(color_box)

    # This function converts an MPL color to a QColor
    @staticmethod
    def convert_to_qcolor(color):
        """
        Converts a provided matplotlib color `color` to a
        :obj:`~PyQt5.QtGui.QColor` object.

        Parameters
        ----------
        color : str or tuple of length {3, 4}
            The matplotlib color that must be converted.

        Returns
        -------
        qcolor : :obj:`~PyQt5.QtGui.QColor` object
            The instance of the :class:`~PyQt5.QtGui.QColor` class that
            corresponds to the provided `color`.

        """

        # Obtain the RGBA values of an MPL color
        r, g, b, a = to_rgba(color)

        # Convert to Qt RGBA values
        color = QG.QColor(
            round(r*255),
            round(g*255),
            round(b*255),
            round(a*255))

        # Return color
        return(color)

    # This function converts a QColor to an MPL color
    @staticmethod
    def convert_to_mpl_color(qcolor):
        """
        Converts a provided :obj:`~PyQt5.QtGui.QColor` object `color` to a
        matplotlib color.

        Parameters
        ----------
        qcolor : :obj:`~PyQt5.QtGui.QColor` object
            The instance of the :class:`~PyQt5.QtGui.QColor` class must be
            converted to a matplotlib color.

        Returns
        -------
        color : str
            The corresponding matplotlib color.
            The returned `color` is always written in HEX.

        """

        hexid = qcolor.name()
        return(str(hexid))

    # This function creates a pixmap of an MPL color
    @staticmethod
    def create_color_pixmap(color, size):
        """
        Creates a :obj:`~PyQt5.QtGui.QPixmap` object consisting of the given
        `color` with the provided `size`.

        Parameters
        ----------
        color : str
            The matplotlib color that must be used for the pixmap.
        size : tuple
            The width and height dimension values of the pixmap to be created.

        Returns
        -------
        pixmap : :obj:`~PyQt5.QtGui.QPixmap` object
            The instance of the :class:`~PyQt5.QtGui.QPixmap` class that was
            created from the provided `color` and `size`.

        """

        # Obtain the RGBA values of an MPL color
        color = ColorBox.convert_to_qcolor(color)

        # Create an image object
        image = QG.QImage(*size, QG.QImage.Format_RGB32)

        # Fill the entire image with the same color
        image.fill(color)

        # Convert the image to a pixmap
        pixmap = QG.QPixmap.fromImage(image)

        # Return the pixmap
        return(pixmap)

    # This function shows the custom color picker dialog
    @QC.Slot()
    def show_colorpicker(self):
        """
        Shows the colorwheel picker dialog to the user, allowing for any color
        option to be selected.

        """

        # Obtain current qcolor
        qcolor = self.convert_to_qcolor(self.get_box_value())

        # Show color dialog
        color = QW.QColorDialog.getColor(
            qcolor, parent=self,
            options=QW.QColorDialog.DontUseNativeDialog)

        # If the returned color is valid, save it
        if color.isValid():
            self._set_color(self.convert_to_mpl_color(color))

    # This function is called whenever the user changes the lineedit
    @QC.Slot(str)
    def set_color(self, color):
        """
        Sets the current color to the provided `color`, and updates the entry
        in the combobox and the label accordingly.

        Parameters
        ----------
        color : str
            The color that needs to be used as the current color. The provided
            `color` can be any string that is accepted as a color by
            matplotlib.
            If `color` is invalid, the default color is used instead.

        """

        # Check if the combobox currently holds an acceptable input
        status = self.color_combobox.lineEdit().hasAcceptableInput()

        # Check status
        if status:
            # If valid, add a hash if color is a 6-digit hex string
            color = re.sub(r"^[\da-fA-F]{6}$", lambda x: '#'+x[0], color)
            set_box_value(self.color_combobox, color)
        else:
            # Else, use the default color
            color = self.default_color

            # If combobox currently has no focus, set combobox value as well
            if not self.color_combobox.hasFocus():
                set_box_value(self.color_combobox, color)

        # Set the color label of the colorbox
        self.set_color_label(color)

    # This function sets a given color as the current color
    @QC.Slot(str)
    def _set_color(self, color):
        # Set the color label
        self.set_color_label(color)

        # Set the combobox to the proper value as well
        set_box_value(self.color_combobox, color)

    # This function sets the color of the colorlabel
    @QC.Slot(str)
    def set_color_label(self, color):
        """
        Sets the current color label to the provided `color`.

        Parameters
        ----------
        color : str
            The color that needs to be used as the current color label. The
            provided `color` can be any string that is accepted as a color by
            matplotlib.

        """

        # Create pixmap of given color
        pixmap = self.create_color_pixmap(color, self.clabel_size)

        # Set the colorlabel
        set_box_value(self.color_label, pixmap)

    # This function sets the default color of the color box
    @QC.Slot()
    @QC.Slot(str)
    def set_default_color(self, color=None):
        """
        Sets the default color value to `color`.

        Optional
        --------
        color : str or None. Default: None
            The matplotlib color value that must be set as the default value
            for this colorbox.
            If *None*, use the current color value of this colorbox instead.

        """

        # If color is None, obtain current value
        if color is None:
            color = self.get_box_value()

        # Set new default color
        self.default_color = color
        self.color_combobox.lineEdit().setPlaceholderText(color)

    # This function retrieves a value of this special box
    def get_box_value(self, *value_sig):
        """
        Returns the current color value of the color combobox.

        Returns
        -------
        color : str
            The current matplotlib color value.

        """

        # Return the value currently set
        return(get_box_value(self.color_combobox, *value_sig))

    # This function sets the value of this special box
    def set_box_value(self, value, *value_sig):
        """
        Sets the current color value to `value`.
        This also sets the default color to `value`.

        Parameters
        ----------
        value : str
            The matplotlib color value that must be set for this colorbox.

        """
        # Convert color to a proper MPL color if it is a float
        if '.' in value:
            value = to_hex(value)

        # Set the current default color
        self.set_default_color(value)
        self._set_color(value)


# Make class with a special box for setting the colormap of a plotted 2D hist
class ColorMapBox(GW.BaseBox):
    """
    Defines the :class:`~ColorMapBox` class.

    This widget allows for colormaps in *matplotlib* to be easily picked by the
    user.
    All colormaps that are registered in :mod:`matplotlib.cm` before the first
    time that this widget is initialized, are available in the combobox.
    Colormaps can be added to *matplotlib* using the
    :func:`~matplotlib.cm.register_cmap` function.

    The first time this widget is initialized, all icons for all colormaps are
    drawn and stored in memory, which can take a few seconds. Every

    """

    # Signals
    modified = QC.Signal([], [str])

    # Set the size for the colormap previews
    cmap_size = (70, 16)

    # Set flag for first_init
    init_flag = False

    # Set property for bad_cmaps
    bad_cmaps = {'gist_ncar', 'gist_rainbow', 'gist_stern', 'hsv', 'jet',
                 'nipy_spectral'}

    # Initialize ColorMapBox
    def __init__(self, parent=None):
        """
        Initialize an instance of the :class:`~ColorMapBox` class.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the colormap box
        self.init()

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[str])

    # This function creates a combobox with colormaps
    def init(self):
        # Check if this class has been initialized before, and do so if not
        if not self.init_flag:
            self.first_init()

        # Create a layout for this widget
        box_layout = GL.QHBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create a combobox for cmaps
        cmaps_box = GW.EditableComboBox()
        validator = GW.ComboBoxValidator(cmaps_box)
        cmaps_box.setValidator(validator)

        # Add all colormaps to cmaps_box
        for cmap in self.cmaps_cl:
            cmap_icon = self.cmap_icons[cmap]
            cmaps_box.addItem(cmap_icon, cmap)

        # Add some separators
        for i in reversed(self.cum_len[:-2]):
            cmaps_box.insertSeparator(i)

        # Set remaining properties
        set_box_value(cmaps_box, rcParams['image.cmap'])
        cmaps_box.setIconSize(QC.QSize(*self.cmap_size))
        cmaps_box.completer().popup().setIconSize(QC.QSize(*self.cmap_size))
        get_modified_signal(cmaps_box, str).connect(self.cmap_selected)
        cmaps_box.focusLost.connect(
            lambda: set_box_value(cmaps_box, get_box_value(cmaps_box, int)))

        # Add cmaps_box to layout
        box_layout.addWidget(cmaps_box)
        self.cmaps_box = cmaps_box

    # This function prepares the class for being initialized for the first time
    def first_init(self):
        # Obtain all colormaps that are registered in MPL
        cmaps = plt.colormaps()

        # Split cmaps up into their cmap types
        cm_types = ['sequential', 'diverging', 'cyclic', 'qualitative', 'misc']
        cmaps_cd = {cm_type: sset() for cm_type in cm_types}
        for cmap in cmaps:
            cmaps_cd[get_cmap_type(cmap)].add(cmap)

        # Create empty list of cmaps sorted on type
        cmaps_cl = []
        cum_len = []

        # Loop over every type
        for cmaps_cs in cmaps_cd.values():
            # Take all base versions of the colormaps
            cmaps_cl.extend([cmap for cmap in cmaps_cs
                             if not cmap.endswith('_r')])
            cum_len.extend([len(cmaps_cl)])

            # Also add all the reversed versions
            cmaps_cl.extend([cmap for cmap in cmaps_cs if cmap.endswith('_r')])
            cum_len.extend([len(cmaps_cl)]*2)

        # Store list of colormaps and the category lengths
        ColorMapBox.cmaps_cl = cmaps_cl
        ColorMapBox.cum_len = cum_len

        # Create the colormap icons
        cmap_icons = sdict()
        for cmap in cmaps:
            cmap_icons[cmap] = self.create_cmap_icon(cmap, self.cmap_size)
        ColorMapBox.cmap_icons = cmap_icons

        # Save that class has been initialized for the first time
        ColorMapBox.init_flag = True

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        self.modified[str].emit(get_box_value(self.cmaps_box))

    # This function creates an icon of a colormap
    @staticmethod
    def create_cmap_icon(cmap, size):
        """
        Creates a :obj:`~PyQt5.QtGui.QIcon` object of the given `cmap` with the
        provided `size`.

        Parameters
        ----------
        cmap : :obj:`~matplotlib.colors.Colormap` object or str
            The colormap for which an icon needs to be created.
        size : tuple
            A tuple containing the width and height dimension values of the
            icon to be created.

        Returns
        -------
        icon : :obj:`~PyQt5.QtGui.QIcon` object
            The instance of the :class:`~PyQt5.QtGui.QIcon` class that was
            created from the provided `cmap` and `size`.

        """

        # Obtain the cmap
        cmap = cm.get_cmap(cmap)

        # Obtain the RGBA values of the colormap
        mplRGBA = cmap(np.arange(cmap.N))

        # Convert to Qt RGBA values
        qtRGBA = [ColorBox.convert_to_qcolor(RGBA).rgba() for RGBA in mplRGBA]

        # Create an image object
        image = QG.QImage(cmap.N, 1, QG.QImage.Format_RGB32)

        # Set the value of every pixel in this image
        for i, RGBA in enumerate(qtRGBA):
            image.setPixel(i, 0, RGBA)

        # Scale the image to its proper size
        image = image.scaled(*size)

        # Convert the image to a pixmap
        pixmap = QG.QPixmap.fromImage(image)

        # Convert the pixmap to an icon
        icon = QG.QIcon(pixmap)

        # Return the icon
        return(icon)

    # This function allows cmaps to be added to the 'bad_cmaps' set
    def addBadCmaps(self, cmaps):
        """
        Adds the provided list of `cmaps` to the set of colormaps that should
        not be used by the user under any circumstances.

        If a user selects a colormap that is considered 'bad', a warning
        message will be shown to the user.

        Parameters
        ----------
        cmaps : list of str
            List of names of colormaps that are registered in
            :mod:`matplotlib.cm` that should be added to the list of 'bad'
            colormaps.

        """

        # Obtain the names of all colormaps
        cmap_names = plt.colormaps()

        # Make sure every colormap provided is a valid colormap
        for cmap in cmaps:
            if cmap in cmap_names and not cmap.endswith('_r'):
                # If so, add it to the set
                self.bad_cmaps.add(cmap)

    # This function checks a selected cmap
    @QC.Slot(str)
    def cmap_selected(self, cmap):
        """
        Qt slot that checks a provided `cmap` and shows an error message if
        `cmap` is a terrible colormap.

        """

        # If a terrible colormap is selected, show error message
        if cmap.startswith(tuple(self.bad_cmaps)):
            # Create error message
            err_msg = ("The selected <b><i>%s</i></b> cmap is a bad choice for"
                       " plotting data. To avoid introducing fake perceptual "
                       "features, it is recommended to pick a <i>perceptually "
                       "uniform sequential</i> colormap, like the ones with "
                       "the <i>cmr.</i> prefix.<br><br>"
                       "See <a href=\"%s\">here</a> for more information on "
                       "this subject." %
                       (cmap,
                        "https://cmasher.readthedocs.io/user/background.html"))

            # Show error window
            QW.QMessageBox.warning(
                self, "%s WARNING" % (cmap.upper()), err_msg)

    # This function retrieves a value of this special box
    def get_box_value(self, *value_sig):
        """
        Returns the current colormap of the colormap box.

        Returns
        -------
        cmap : str or :obj:`~matplotlib.colors.Colormap` object
            The currently selected colormap.

        """

        # Obtain the value
        cmap = get_box_value(self.cmaps_box)

        # Obtain the Colormap object if requested
        if Colormap in value_sig:
            cmap = plt.get_cmap(cmap)

        # Return it
        return(cmap)

    # This function sets the value of this special box
    def set_box_value(self, value, *value_sig):
        """
        Sets the current colormap to `value`.

        Parameters
        ----------
        value : str or :obj:`~matplotlib.colors.Colormap` object
            The colormap that must be used for this colormap box.

        """

        # Obtain the name of the provided colormap if needed
        if isinstance(value, Colormap):
            value = value.name

        # Set this as the current colormap
        set_box_value(self.cmaps_box, value)



# -*- coding: utf-8 -*-

"""
Exception Widgets
=================
Provides all definitions required for handling exceptions in *GuiPy*.

"""


# %% IMPORTS
# Built-in imports
from traceback import format_exception_only, format_tb

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW

# All declaration
__all__ = ['ExceptionDialog', 'create_exception_handler']


# %% CLASS DEFINITIONS
# Make special class for showing exception details
class ExceptionDialog(GW.QDialog):
    """
    Defines the :class:`~ExceptionDialog` class.

    This class takes a set of exception details and converts it into a format
    that can be shown using a dialog.

    """

    def __init__(self, etype, value, tb, parent=None):
        """
        Initialize an instance of the :class:`~ExceptionDialog` class.

        Parameters
        ----------
        etype : :class:`~Exception` class
            The :class:`~Exception` class that is associated with this error.
        value : :obj:`~Exception` object
            The :class:`~Exception` instance that is associated with this
            error.
        tb : traceback object
            The corresponding traceback object.

        Optional
        --------
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget for this dialog or *None* for no parent.

        """

        # Save the provided values
        self.etype = etype
        self.value = value
        self.tb = tb

        # Call the super constructor
        super().__init__(parent)

        # Initialize the exception dialog
        self.init()

    # This function creates the exception dialog
    def init(self):
        """
        Sets up the exception dialog after it has been initialized.

        This function is mainly responsible for gathering all required
        information; formatting it; and drawing the dialog.

        """

        # Create a window layout
        grid_layout = GL.QGridLayout(self)
        grid_layout.setColumnStretch(2, 1)
        grid_layout.setRowStretch(3, 1)

        # Set properties of message box
        self.setWindowModality(QC.Qt.ApplicationModal)
        self.setAttribute(QC.Qt.WA_DeleteOnClose)
        self.setWindowTitle("ERROR")
        self.setWindowFlags(
            QC.Qt.MSWindowsOwnDC |
            QC.Qt.Dialog |
            QC.Qt.WindowTitleHint |
            QC.Qt.WindowSystemMenuHint |
            QC.Qt.WindowCloseButtonHint)

        # Set the icon of the exception on the left
        icon_label = GW.QLabel()
        pixmap = GW.QMessageBox.standardIcon(GW.QMessageBox.Critical)
        icon_label.setPixmap(pixmap)
        grid_layout.addWidget(icon_label, 0, 0, 2, 1, QC.Qt.AlignTop)

        # Add a spacer item
        spacer_item = QW.QSpacerItem(7, 1, QW.QSizePolicy.Fixed,
                                     QW.QSizePolicy.Fixed)
        grid_layout.addItem(spacer_item, 0, 1, 2, 1)

        # Set the text of the exception
        exc_str = self.format_exception()
        exc_label = GW.QLabel(exc_str)
        grid_layout.addWidget(exc_label, 0, 2, 1, 1)

        # Create a button box for the buttons
        button_box = QW.QDialogButtonBox()
        grid_layout.addWidget(button_box, 2, 0, 1, grid_layout.columnCount())

        # Create traceback box
        self.tb_box = self.create_traceback_box()
        grid_layout.addWidget(self.tb_box, 3, 0, 1, grid_layout.columnCount())

        # Create traceback button
        self.tb_but =\
            button_box.addButton(self.tb_labels[self.tb_box.isHidden()],
                                 button_box.ActionRole)
        self.tb_but.clicked.connect(self.toggle_traceback_box)

        # Create an 'ok' button
        ok_but = button_box.addButton(button_box.Ok)
        ok_but.clicked.connect(self.close)
        ok_but.setDefault(True)

        # Update the size
        self.update_size()

    # This function formats the exception string
    def format_exception(self):
        """
        Formats the exception provided during initialization and returns it.

        """

        # Format the exception
        exc_list = format_exception_only(self.etype, self.value)
        exc_str = ''.join(exc_list)

        # Return it
        return(exc_str)

    # This function formats the traceback string
    def format_traceback(self):
        """
        Formats the traceback provided during initialization and returns it.

        """

        # Format the traceback
        tb_list = format_tb(self.tb)
        tb_str = ''.join(tb_list)

        # Return it
        return(tb_str)

    # This function creates the traceback box
    def create_traceback_box(self):
        """
        Creates a special box for the exception dialog that contains the
        traceback information and returns it.

        """

        # Create a traceback box
        traceback_box = GW.QWidget()
        traceback_box.setHidden(True)

        # Create layout
        layout = GL.QVBoxLayout(traceback_box)
        layout.setContentsMargins(0, 0, 0, 0)

        # Add a horizontal line to the layout
        layout.addSeparator()

        # Format the traceback
        tb_str = self.format_traceback()

        # Add a textedit to the layout
        tb_text_box = GW.QTextEdit(traceback_box)
        tb_text_box.setMinimumHeight(100)
        tb_text_box.setFocusPolicy(QC.Qt.NoFocus)
        tb_text_box.setReadOnly(True)
        tb_text_box.setText(tb_str)
        layout.addWidget(tb_text_box)

        # Create a 'show traceback' button
        self.tb_labels = ['Hide Traceback...', 'Show Traceback...']

        # Return traceback box
        return(traceback_box)

    # This function shows or hides the traceback box
    @QC.Slot()
    def toggle_traceback_box(self):
        """
        Toggles the visibility of the traceback box and updates the dimensions
        of the exception dialog accordingly.

        """

        # Toggle the visibility of the traceback box
        self.tb_box.setVisible(self.tb_box.isHidden())
        self.tb_but.setText(self.tb_labels[self.tb_box.isHidden()])

        # Update the size of the message box
        self.update_size()

    # This function updates the size of the dialog
    def update_size(self):
        """
        Updates the dimensions of the exception dialog depending on its current
        state (traceback box visibility).

        """

        # Determine the minimum/maximum size required for making the dialog
        min_size = self.layout().minimumSize()
        max_size = self.layout().maximumSize()

        # Set the fixed width
        self.setFixedWidth(min_size.width())

        # If the traceback box is shown, set minimum/maximum height
        if self.tb_box.isVisible():
            self.setMinimumHeight(min_size.height())
            self.setMaximumHeight(max_size.height())
        # Else, set fixed height
        else:
            self.setFixedHeight(min_size.height())


# %% FUNCTION DEFINITIONS
# This function factory creates a function for handling exceptions
def create_exception_handler(parent=None):
    """
    Function factory that returns a function definition
    ``show_exception_details(*args, **kwargs)``.

    Calling the returned definition will automatically initialize an
    :obj:`~guipy.widgets.exceptions.ExceptionDialog` object using the provided
    `parent`, `args` and `kwargs`; and show the dialog.

    Optional
    --------
    parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
        The parent widget to use for the exception dialog or *None* for no
        parent.

    """

    # This function creates a message box with exception information
    def show_exception_details(*args, **kwargs):
        """
        Creates an instance of the :class:`~ExceptionDialog` class and shows
        it.

        Parameters
        ----------
        args : positional arguments
            The positional arguments that must be passed to the constructor of
            the :class:`~guipy.widgets.exceptions.ExceptionDialog` class.
        kwargs : keyword arguments
            The keyword arguments that must be passed to the constructor of the
            :class:`~guipy.widgets.exceptions.ExceptionDialog` class.

        """

        # Create exception message box
        exception_box = ExceptionDialog(*args, parent=parent, **kwargs)

        # Emit the exception signal of the parent if it has it
        if hasattr(parent, 'exception'):
            parent.exception.emit()

        # Show the exception message box
        exception_box.show()

    # Return function definition
    return(show_exception_details)



# -*- coding: utf-8 -*-

"""
Marker Property
===============

"""


# %% IMPORTS
# Package imports
from matplotlib import rcParams

# GuiPy imports
from guipy import widgets as GW
from guipy.plugins.figure.widgets.types.props import BasePlotProp
from guipy.widgets import set_box_value

# All declaration
__all__ = ['LineMarkerProp', 'ScatterMarkerProp']


# %% CLASS DEFINITIONS
# Define 'Marker' plot property
class MarkerProp(BasePlotProp):
    # Class attributes
    DISPLAY_NAME = "Marker"
    WIDGET_NAMES = [*BasePlotProp.WIDGET_NAMES, 'marker_style_box',
                    'marker_size_box', 'marker_color_box']

    # This function creates and returns a line style box
    def marker_style_box(self):
        """
        Creates a widget box for setting the style of the marker and returns
        it.

        """

        # Make combobox for markerstyles
        marker_style_box = GW.MarkerStyleBox()
        marker_style_box.setToolTip("Marker to be used for this plot")

        # Set initial value to the default value in MPL
        set_box_value(marker_style_box, self.default_marker)

        # Return name and box
        return('Style', marker_style_box)

    # This function creates and returns a marker size box
    def marker_size_box(self):
        """
        Creates a widget box for setting the size of the marker and returns it.

        """

        # Make a double spinbox for markersize
        marker_size_box = GW.QDoubleSpinBox()
        marker_size_box.setToolTip("Size of the plotted markers")
        marker_size_box.setRange(0, 9999999)
        marker_size_box.setSuffix(" pts")

        # Set initial value to the default value in MPL
        set_box_value(marker_size_box, rcParams['lines.markersize'])

        # Return name and box
        return('Size', marker_size_box)

    # This function creates and returns a marker color box
    def marker_color_box(self):
        """
        Creates a widget box for setting the color of the marker and returns
        it.

        """

        # Make a color box
        marker_color_box = GW.ColorBox()
        marker_color_box.setToolTip("Color to be used for this marker")

        # Connect 'applying' signal
        self.options.applying.connect(marker_color_box.set_default_color)

        # Return name and box
        return('Color', marker_color_box)


# Define 'LineMarker' plot property
class LineMarkerProp(MarkerProp):
    """
    Provides the definition of the :class:`~LineMarkerProp` plot property,
    specifically used for line plots.

    This property contains boxes for setting the marker style, marker size and
    marker color.

    """

    # Class attributes
    NAME = "LineMarker"

    # This property holds the default marker used for line plots
    @property
    def default_marker(self):
        return(rcParams['lines.marker'])


# Define 'ScatterMarker' plot property
class ScatterMarkerProp(MarkerProp):
    """
    Provides the definition of the :class:`~ScatterMarkerProp` plot property,
    specifically used for scatter plots.

    This property contains boxes for setting the marker style, marker size and
    marker color.

    """

    # Class attributes
    NAME = "ScatterMarker"

    # This property holds the default marker used for scatter plots
    @property
    def default_marker(self):
        return(rcParams['scatter.marker'])



# -*- coding: utf-8 -*-

"""
Text
====
Provides a collection of :class:`~PyQt5.QtWidgets.QWidget` subclasses for
handling text and labels in :mod:`~matplotlib`.

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtWidgets as QW

# GuiPy imports
from guipy import layouts as GL, widgets as GW
from guipy.widgets import get_box_value, set_box_value
from guipy.widgets.spinbox import FontSizeBox

# All declaration
__all__ = ['FigureLabelBox']


# %% CLASS DEFINITIONS
# Make class for setting the label in a figure
class FigureLabelBox(GW.DualBaseBox):
    """
    Defines the :class:`~FigureLabelBox` class.

    """

    # Signals
    modified = QC.Signal([], [str], [str, dict])

    # Initialize the FigureLabelBox class
    def __init__(self, parent=None):
        """
        Initialize an instance of the :class:`~FigureLabelBox` class.

        Optional
        --------
        parent : :obj:`~PyQt5.QtWidgets.QWidget` object or None. Default: None
            The parent widget to use for this figure label box or *None* for no
            parent.

        """

        # Call super constructor
        super().__init__(parent)

        # Create the figure label box
        self.init()

    # This property returns the default 'modified' signal
    @property
    def default_modified_signal(self):
        return(self.modified[str])

    # This function creates the figure label box
    def init(self):
        """
        Sets up the figure label box after it has been initialized.

        """

        # Create the box_layout
        box_layout = GL.QHBoxLayout(self)
        box_layout.setContentsMargins(0, 0, 0, 0)

        # Create a line-edit for setting the label
        label_box = GW.QLineEdit()
        box_layout.addWidget(label_box)
        self.left_box = label_box

        # Create a spinbox for setting the fontsize
        size_box = FontSizeBox()
        box_layout.addWidget(size_box)
        self.right_box = size_box

    # This function is automatically called whenever 'modified' is emitted
    @QC.Slot()
    def modified_signal_slot(self):
        # Emit modified signals
        self.modified[str].emit(get_box_value(self.left_box))
        self.modified[str, dict].emit(*self.get_box_value())

    # This function retrieves a value of this special box
    def get_box_value(self, *value_sig):
        """
        Returns the current values of this figure label box as a tuple.

        Returns
        -------
        value : tuple
            A tuple containing the values of the figure label box, formatted as
            `(label, {'fontsize': size})`.

        """

        return(get_box_value(self.left_box),
               {'fontsize': get_box_value(self.right_box)})

    # This function sets the value of this special box
    def set_box_value(self, value, *value_sig):
        """
        Sets the current value of the figure label box to `value`.

        Parameters
        ----------
        value : tuple
            A tuple containing the values of the figure label box, formatted as
            `(label, {'fontsize': size})`.

        """

        # Set box value of label and size
        set_box_value(self.left_box, value[0])
        set_box_value(self.right_box, value[1]['fontsize'])



# -*- coding: utf-8 -*-

"""
Configuration
=============
Contains all the configuration files and functions of *GuiPy*.

"""


# %% IMPORTS
# Import core modules
from . import core, dialog, manager
from .core import *
from .dialog import *
from .manager import *
from . import base
from .base import *

# Import config pages
from . import general
from .general import *

# All declaration
__all__ = ['base', 'core', 'dialog', 'general', 'manager']
__all__.extend(base.__all__)
__all__.extend(core.__all__)
__all__.extend(dialog.__all__)
__all__.extend(general.__all__)
__all__.extend(manager.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"



# -*- coding: utf-8 -*-

"""
Data Table Headers
==================

"""


# %% IMPORTS
# Built-in imports

# Package imports
from qtpy import QtCore as QC, QtGui as QG, QtWidgets as QW

# GuiPy imports
from guipy import widgets as GW

# All declaration
__all__ = ['BaseHeaderView', 'HorizontalHeaderView', 'VerticalHeaderView']


# %% CLASS DEFINITIONS
# Define base class for the headers of the data table
class BaseHeaderView(GW.QHeaderView):
    # Signal
    sectionRightClicked = QC.Signal(int)

    # Initialize BaseHeaderView class
    def __init__(self, orientation, parent):
        # Call super constructor
        super().__init__(orientation, parent)

    # This function sets up the header view
    def init(self, context_menu, double_clicked=None):
        # Set up default properties
        # TODO: Figure out how to avoid selecting all non-visible items when
        # column/row is selected
#        self.setSectionsClickable(True)
        self.setHighlightSections(True)
        self.setDropIndicatorShown(True)
        self.setContextMenuPolicy(QC.Qt.CustomContextMenu)

        # Set signal handling
        self.customContextMenuRequested.connect(context_menu)
        if double_clicked is not None:
            self.sectionDoubleClicked.connect(double_clicked)

    # Override mousePressEvent to emit a signal when section is right-clicked
    def mousePressEvent(self, event):
        # If event is right clicking, emit sectionRightClicked signal
        if(event.button() == QC.Qt.RightButton):
            self.sectionRightClicked.emit(self.logicalIndexAt(event.pos()))
        return(super().mousePressEvent(event))


# Horizontal data table header
class HorizontalHeaderView(BaseHeaderView):
    # Initialize HorizontalHeaderView class
    def __init__(self, parent, *args, **kwargs):
        # Call super constructor
        super().__init__(QC.Qt.Horizontal, parent)

        # Set up horizontal header view
        self.init(*args, **kwargs)

    # This function sets up the horizontal header view
    def init(self, *args, **kwargs):
        # Call super init
        super().init(*args, **kwargs)

        # Set up default properties
        self.setSortIndicatorShown(True)
#        self.setSectionsMovable(True)
        self.setSectionResizeMode(self.Interactive)


# Vertical data table header
class VerticalHeaderView(BaseHeaderView):
    # Initialize VerticalHeaderView class
    def __init__(self, parent, *args, **kwargs):
        # Call super constructor
        super().__init__(QC.Qt.Vertical, parent)

        # Set up vertical header view
        self.init(*args, **kwargs)

    # This function sets up the vertical header view
    def init(self, *args, **kwargs):
        # Call super init
        super().init(*args, **kwargs)

        # Set up default properties
        self.setSectionResizeMode(self.Fixed)



# -*- coding utf-8 -*-

"""
Layouts
=======
Contains the various different :class:`~PyQt5.QtWidgets.QLayout` subclasses
created for *GuiPy*.

"""


# %% IMPORTS
# Import core modules
from . import base
from .base import *

# All declaration
__all__ = ['base']
__all__.extend(base.__all__)

# Author declaration
__author__ = "Ellert van der Velden (@1313e)"
